import time
import sys
import random
# import argparse
import queue
import functools as ft
import threading
import numpy as np
import numpy.typing as npt
import typing
from pathlib import Path
from importlib import resources

import fire
from PySide6 import QtCore, QtWidgets, QtGui, QtCharts
import pyqtgraph as pg
import pythonosc
import pythonosc.osc_server

import tungnaa.gui
from tungnaa.gui.downloads import dl_model

# multiprocessing: need to use spawn (no fork on mac, no forkserver on windows)
# note, this may preclude pyinstaller...

# to convert obj.method(*args) to multiprocessing...
import multiprocessing as mp
class Proxy:
    """
    this is a somewhat of a hack to try running the backend in a separate
    process with otherwise minimal changes.

    wrapping the Backend object in a Proxy allows the frontend to asynchronously call methods on the backend (but they don't return)

    all communication from the backend to the frontend is through a Pipe,
    which works basically the same as it was previously with a Queue.

    it's not possible to read attributes of the backend.

    to me this design is suggesting to use asyncio to allow backend calls to
    return asynchronously when needed, and the frontend to use `await` when convenient.

    IPyC (https://pypi.org/project/IPyC/) might be helpful there.

    alternatively, Futures as in concurrent.futures (https://docs.python.org/3/library/concurrent.futures.htm) could be useful -- but I'm unsure whether ProcessPoolExecutor is a good fit for calling methods of one long-lived object in another process.

    However, this pattern might turn out to work fine -- the frontend 
    passes data/triggers to the backend asynchronously, and the backend
    streams state updates to the frontend via the Pipe.
    """
    def __init__(self, obj):
        ctx = mp.get_context('spawn')
        # backend object
        self.obj = obj
        # parent and child process ends of the pipe
        self.parent_conn, child_conn = ctx.Pipe()
        # child process eventually sets this to true and sends it through the pipe
        self.running = False
        # run in a new process
        self.proc = ctx.Process(target=self._run, args=(child_conn,))
        self.proc.start()

    def __getattr__(self, name):
        # print(self.__dict__)
        if name in self.__dict__:
            return self.__dict__[name]
        elif len(self.__dict__):
            if not hasattr(self.obj, name):
                # this is checking the parent process copy of obj, meaning
                # it can only find attributes which exist before the fork
                raise AttributeError
            
            if not self.running:
                # first value received from the backend process says it has started
                # if self.parent_conn.poll():
                try:
                    assert self.parent_conn.recv()
                except AssertionError:
                    print("""backend process failed to start""")
                    raise
                self.running = True
                # else:
                    # return lambda *a, **kw: print(f'skipping {name} (backend not running yet)')
            
            attr = getattr(self.obj, name)
            if callable(attr):
                # return a function which when called, defers onto the backend process
                def f(*a, **kw):
                    # print(name)
                    self.parent_conn.send([name, a, kw])
                return f
            else:
                print("""backend attribute not callable""")
            
            raise NotImplementedError
        else:
            # this case allows unpickling
            raise AttributeError
    
    def _run(self, conn):
        print('======================BACKEND RUN======================')
        self.obj.run(conn)
        self.running = True
        conn.send(self.running)

        while True:
            # call function in backend process asynchronously (no return)
            name, a, kw = conn.recv()
            # print(name)
            getattr(self.obj, name)(*a, **kw)


# save some boilerplate
def Layout(*items, cls=QtWidgets.QHBoxLayout, spacing=None, margins=None):
    l = cls()
    if spacing is not None:
        l.setSpacing(spacing)
    if margins is not None:
        l.setContentsMargins(*margins)
    for item in items:
        if isinstance(item, QtWidgets.QWidget):
            l.addWidget(item)
        elif isinstance(item, QtWidgets.QLayout):
            l.addLayout(item)
    return l
HBoxLayout = ft.partial(Layout, cls=QtWidgets.QHBoxLayout)
VBoxLayout = ft.partial(Layout, cls=QtWidgets.QVBoxLayout)

class DoubleSlider(QtWidgets.QSlider):
    """
    Continuous value QSlider that uses native double values instead of integers
    """

    # define a signal which enables Qt's callback system of signals/slots
    doubleValueChanged = QtCore.Signal(float)

    def __init__(self, decimals=3, *args, **kargs):
        super(DoubleSlider, self).__init__( *args, **kargs)
        self._multi = 10 ** decimals
        self.valueChanged.connect(self.emitDoubleValueChanged)

    def emitDoubleValueChanged(self):
        value = float(super(DoubleSlider, self).value())/self._multi
        self.doubleValueChanged.emit(value)

    def value(self):
        return float(super(DoubleSlider, self).value()) / self._multi

    def setMinimum(self, value):
        return super(DoubleSlider, self).setMinimum(value * self._multi)

    def setMaximum(self, value):
        return super(DoubleSlider, self).setMaximum(value * self._multi)

    def setSingleStep(self, value):
        return super(DoubleSlider, self).setSingleStep(value * self._multi)

    def singleStep(self):
        return float(super(DoubleSlider, self).singleStep()) / self._multi

    def setValue(self, value:float):
        super(DoubleSlider, self).setValue(int(value * self._multi))


class PlotWidget(pg.PlotWidget):
    # disable zoom
    def wheelEvent(self, ev):
        pass

class AlignmentImageItem(pg.ImageItem):
    def __init__(self, clicked):
        super().__init__()
        self.clicked = clicked
    def mousePressEvent(self, ev):
        if self.clicked is not None:
            self.clicked(ev)

class AlignmentGraph(QtWidgets.QWidget):
    """
    Widget for displaying alignments
    """
    def __init__(self, 
        parent:QtWidgets.QWidget=None, 
        image_clicked:typing.Callable=None):
        super().__init__(parent)

        self.max_display_steps = 200

        pg.setConfigOption('imageAxisOrder', 'row-major') # best performance image data must be (height, width)
        #pg.setConfigOption('useNumba', True) # supposedly better performance for image data

        self.num_encodings = 40

        self.frame = 0
        self.prev_tok = 0

        self.plot_widget = pg.GraphicsLayoutWidget(self)
        self.plot_widget.ci.layout.setContentsMargins(0,0,0,4)
        self.plot_widget.ci.layout.setSpacing(0)
        self.plot = self.plot_widget.addPlot()

        # self.plot = PlotWidget(parent=self)
        # See: https://pyqtgraph.readthedocs.io/en/latest/api_reference/graphicsItems/imageitem.html#pyqtgraph.ImageItem
        self.imageitem = AlignmentImageItem(image_clicked)
        # self.imageitem.setImage(image=self.imagedata.T)
        self.plot.addItem(self.imageitem)
        self.plot.showAxes(True, showValues=(True, True, True, False))
        self.plot.invertY(False) # vertical axis zero at the bottom
        # no idea why, but fully transparent background is too light
        # black with alpha of 40 seems to match
        self.plot_widget.setBackground((0,0,0,40))

        self.attn_slider_resolution = 512
        self.attn_slider_max_value = (
            self.num_encodings-1) * self.attn_slider_resolution
        self.alignment_slider = QtWidgets.QSlider(
            parent=self, orientation=QtCore.Qt.Horizontal)
        self.alignment_slider.setMinimum(0)
        self.alignment_slider.setMaximum( self.attn_slider_max_value )

        self.layout = VBoxLayout(
            self.alignment_slider, self.plot_widget, 
            spacing=0, margins=(0,0,0,0))
        self.setLayout(self.layout)

        self.text = None


    def set_normalized_alignment(self, value:float):
        """
        Sets alignment slider to a normalized position from 0-1

        TODO: There's some kind of bug here when controlling via OSC, where the /set_alignment command doesn't scale properly from 0-1 to the range of tokens
                my guess is that self.attn_slider_max_value isn't getting updated? Or something else weird is going on...
        """
        val_as_token = int(value*self.attn_slider_max_value)
        self.alignment_slider.setValue(val_as_token)
        print(f"Set normalized alignment {value} - as token: {val_as_token}/{self.attn_slider_max_value}/{self.num_encodings} ")

    def set_alignment_as_token_idx(self, tok:int) -> None:
        """
        Set alignment slider to a position corresponding to a given token index
        between 0 and self.num_encodings-1

        tok     the target token index (must be between 0 and self.num_tokens-1)
        """

        if tok >= 0 and tok < self.num_encodings:
            tok_as_sliderval = tok * self.attn_slider_resolution
            self.alignment_slider.setValue(tok_as_sliderval)
        else:
            print(f"Error: token index {tok} out of range (0-{self.num_encodings-1})")

    def get_slidervalue_as_params(self) -> np.ndarray:
        """slider value to attention parameters"""
        tok = self.alignment_slider.value() / self.attn_slider_resolution 

        # smoothing:
        tok = (tok + self.prev_tok) / 2
        width = abs(tok - self.prev_tok)
        self.prev_tok = tok
    
        return tok, width

    def addFrame(self, newframe:npt.ArrayLike, text:str):
        # TODO: discard frames which are out of display range
        # use a circular buffer instead of np.append
        if text != self.text:
            self.set_text(text)
        self.frame +=1
        start_step = max(0, self.frame-1-self.max_display_steps)
        self.imagedata = np.append(self.imagedata, newframe, axis=0)
        self.imageitem.setImage(image=self.imagedata)
        self.imageitem.update()
        self.plot.setYRange(start_step, self.frame-1, padding=0)

    def set_text(self, text):
        self.text = text

        ticks = [(i+0.5,c) for i,c in enumerate(text)]
        # print(ticks)
        self.plot.getAxis('top').setTicks([ticks, []])
        self.plot.getAxis('top').showLabel()
        # self.plot.getAxis('bottom').showLabel(True)
        self.plot.setXRange(0, len(text), padding=0)

    def reset(self, text:str=None):
        """
        Reset the attention graph with a given number of encoded tokens (y-axis)
        Usually called after new text is encoded / attention recalculated.

        num_encodings sets the number of encoded tokens to scale the y-axis by
            if not set the current number of encodings is left as-is
        """
        print(f"AlignmentGraph: reset")
        self.frame = 1
        if text is not None:
            self.num_encodings = len(text)
            self.set_text(text)
        self.attn_slider_max_value = (
            self.num_encodings * self.attn_slider_resolution)
        self.alignment_slider.setMinimum(0)
        self.alignment_slider.setMaximum(self.attn_slider_max_value)
        self.imagedata = np.zeros((1,self.num_encodings), dtype=np.float32)
        # self.imageitem.setImage(image=self.imagedata.T)
        self.plot.setYRange(0, 1, padding=0)
        self.imageitem.setImage(image=self.imagedata)
        self.imageitem.update()
        #self.plot.update()
        

class RaveLatents(QtWidgets.QWidget):
    def __init__(self, parent:QtWidgets.QWidget=None):
        super().__init__(parent)

        self.latents = list()
        self.layout = QtWidgets.QHBoxLayout(self)
        self.is_init = False

    def _init(self, num_latents, pitch_slider):
        for idx in range(num_latents):
            if idx==0 and pitch_slider:
                bmin, bmax = -100., 100.
                vmin, vmax = 50., 550.
            else:
                bmin, bmax = -3., 3.
                vmin, vmax = -5., 5.

            # latent widgets, each is a (SliderWidget, MeterWidget) pair
            # the first slide is bias, second is a display
            bias_slider = DoubleSlider(decimals=3, parent=self)
            bias_slider.setMaximum(bmax)
            bias_slider.setMinimum(bmin)
            bias_slider.setValue(0.0)
            bias_slider.doubleValueChanged.connect(
                lambda val,latent=idx: self._bias_adjust(val, latent))
            bias_slider.setStatusTip(f"bias latent dimension {idx}")
            
            value_meter = DoubleSlider(decimals=3, parent=self)
            value_meter.setMaximum(vmax)
            value_meter.setMinimum(vmin)
            value_meter.setValue(0.0)
            value_meter.setStyleSheet("""
                QSlider::groove:vertical {
                    background-color: black;
                    width: 10px;
                    border-radius: 5;
                }
                                      
                QSlider::add-page:vertical {
                    background: pink;
                    border-radius: 5;
                }
                QSlider::sub-page:vertical {
                    background: black;
                    border-radius: 5;
                }
                """)

            self.latents.append((bias_slider, value_meter))
            self.layout.addWidget(bias_slider)
            self.layout.addWidget(value_meter)
            if idx < num_latents-1:
                self.layout.addSpacing(12)
            self.layout.setSpacing(4)

        self.setMaximumHeight(200)

        self.is_init = True

    def _bias_adjust(self, val:float, latent:int):            
        # NOTE: The bias values of the sliders get sampled regularly in `update`, they are read by get_biases 
        print(f"ADJUST BIAS of LATENT {latent} = {val}")

    def get_bias(self, latent:int) -> float:
        return self.latents[latent][0].value()

    def get_biases(self) -> typing.List[float]:
    # def get_biases(self) -> npt.NDArray[np.float32]:
        # res = np.zeros((1,len(self.latents)), dtype=np.float32)
        # for idx, (bias, _) in enumerate(self.latents):
            # res[0][idx] = bias.value()
        # return res
        return [bias.value() for bias,_ in self.latents]
    
    def set_latent_bias(self, latent:int, value:float):
        """
        Set a latent bias value in the GUI
        """
        self.latents[latent][0].setValue(value)

    def add_latent_bias(self, latent:int, value:float):
        """
        Add a small value to the current bias value of a latent in the GUI
        """
        newval = self.latents[latent][0].value() + value;
        self.latents[latent][0].setValue(newval)

    def reset_latent_biases(self, value:float=0.0):
        """
        Reset all latent biases to 0.0
        """
        for bias,_ in self.latents:
            bias.setValue(value);

    def set_latents(self, values:typing.Union[list, npt.ArrayLike]):
        """
        Set latent values in the gui. 
        values > the number of sliders are ignored

        values  latent values as floats, in the shape (1, num_latents)
        """
        values = values[0] # trim off the extra dimension
        if len(values) <= len(self.latents):
            for idx,val in enumerate(values):
                self.latents[idx][1].setValue(val)
        else: # More values than sliders
            for idx,sliders in enumerate(self.latents):
                sliders[1].setValue(values[idx])


class SettingsDialog(QtWidgets.QDialog):
    def __init__(self, parent=None, context:'MainWindow'=None) -> None:
        super().__init__(parent)
        self.setWindowTitle("Settings")

        def osc_listen_button_callback(val):
            result=None
            addr = (self._oschostline.text(), int(self._oscportline.text()))
            context.osc_listen_addr = addr
            status = self._osclistenbut.isChecked()
            if status:
                self._osclistenbut.setStyleSheet(
                    "color: black; background-color: lightblue")
                self._osclistenbut.setText("Listening..")
                # TODO: This use of MainWindow as context feels kludgy, should use an abstract class / mixin?
                if hasattr(context, "serve_osc"):
                    result = context.serve_osc(address=addr)
                else:
                    print("ERROR: No method 'serve_osc' defined on settings context.")
            else:
                self._osclistenbut.setStyleSheet(
                    "color: black; background-color: grey")
                self._osclistenbut.setText("Listen")
                if hasattr(context, "unserve_osc"):
                    result = context.unserve_osc()
                else:
                    print("ERROR: No method 'serve_osc' defined on settings context.")

            print(f"Listen button status: {status}")

        # Create GUI
        self._info = QtWidgets.QLabel("Settings...")

        self._osc = QtWidgets.QGroupBox(title="OSC Control Setup", parent=self)
        self._oschostline = QtWidgets.QLineEdit(context.osc_listen_addr[0])
        self._oschostline.setMaxLength(17)
        self._oscportline = QtWidgets.QLineEdit(str(context.osc_listen_addr[1]))
        self._oscportline.setMaxLength(5)
        self._osclistenbut = QtWidgets.QPushButton("&Listen")
        self._osclistenbut.setCheckable(True)
        self._osclistenbut.clicked.connect(osc_listen_button_callback)
        self._oscinfolabel = QtWidgets.QLabel("")
        hostportlayout = QtWidgets.QHBoxLayout()
        hostportlayout.addWidget(self._oschostline)
        hostportlayout.addWidget(QtWidgets.QLabel(":"))
        hostportlayout.addWidget(self._oscportline)
        hostport = QtWidgets.QWidget()
        hostport.setLayout(hostportlayout)
        osclayout = QtWidgets.QVBoxLayout()
        osclayout.addWidget(hostport)
        osclayout.addWidget(self._osclistenbut)
        osclayout.addWidget(self._oscinfolabel)
        self._osc.setLayout(osclayout)
        
        OKCLOSE = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self._buttonBox = QtWidgets.QDialogButtonBox(OKCLOSE)
        self._buttonBox.accepted.connect(self.accept)
        self._buttonBox.rejected.connect(self.reject)

        self._oscinstructions = QtWidgets.QLabel(context.osc_controller.print_osc_api())
        self._oscinstructions.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        self._oscinstructions_scroller = QtWidgets.QScrollArea()
        self._oscinstructions_scroller.setWidget(self._oscinstructions)

        self._main_layout = VBoxLayout(
            self._info,
            self._osc,
            self._buttonBox,
            QtWidgets.QFrame(),
            self._oscinstructions_scroller
        )

        self.setLayout(self._main_layout)

# Glue between independent threads and the main Qt GUI
class Signaller(QtCore.QObject) :
    # List signals here
    set_gen_text = QtCore.Signal(str)
    set_samp_text = QtCore.Signal(str)

signaller = Signaller()

# Controller thread that receives OSC messages and maps them to Tungnaa actions
class OSCController(threading.Thread):

    def __init__(self, context:'MainWindow', *args, **kwargs):
        super(OSCController, self).__init__(*args, **kwargs)
        self.context = context
        self.osc_server = None
        self.osc_dispatcher = None    

        # BEGIN: OSC Dispatcher Callback Functions -----------------------------
        def osc_unknown(addr:str, *args:list[typing.Any]) -> None:
            print(f"Unknown OSC address: {addr}  with: '{args}'")

        def osc_generate(addr:str, reset:bool=True) -> None:
            """Run/Play generator mode
            reset   bool if true reset to beginning of utterance before playback (useful if generate_stop_at_end is active)
            """
            # TODO: maybe adds some timing wierdness having to go through the QtGui thread for the reset..?
            if reset:
                self.context._reset_action.trigger()
            self.context._generate_action.trigger()

        def osc_sampler(addr:str, *args:list[typing.Any]) -> None:
            """(trigger)Activate sampler mode"""
            self.context._sampler_action.trigger()

        def osc_pause(addr:str, *args:list[typing.Any]) -> None:
            """(trigger)Pause generator/sampler playback"""
            self.context._pause_action.trigger()

        def osc_reset(addr:str, *args:list[typing.Any]) -> None:
            """(trigger)Reset generator/sampler playback"""
            self.context._reset_action.trigger()

        def osc_set_bias(addr:str, latent:int, bias:float) -> None:
            """Set vocoder latent bias
            latent      int index of latent
            bias        float bias amount (usually not larger than +-3.0)

            """
            self.context.latents.set_latent_bias(latent, bias)
            print(f"Set latent {latent} bias: {bias}")

        def osc_add_bias(addr:str, latent:int, bias:float) -> None:
            """Set add a value to the current vocoder latent bias
            latent      int index of latent
            bias        float amount to add to bias (usually a small value, will clip if larger than +-3.0)

            """
            self.context.latents.add_latent_bias(latent, bias)
            print(f"Add {bias} to latent {latent} bias")


        def osc_reset_biases(addr:str) -> None:
            """Reset all biases to 0.0"""
            self.context.latents.reset_latent_biases()
            print(f"Reset latent biases to 0.0")


        def osc_generate_stop_at_end(addr:str, enable:bool) -> None:
            """ Enable/disable Generator stop-at-end of utterance
            enable     bool true or false
            """
            self.context._generate_stop_at_end_toggle_action.setChecked(enable)

        def osc_alignment_mode(addr:str, mode:str) -> None:
            """Set alignment generation mode (Generator only)
            mode    str 'infer' for alignment inference, 'paint' for alignment painting
            """
            self.context.set_alignment_mode(mode)

        def osc_latent_feedback(addr:str, enable:bool) -> None:
            """Enable/disable latent feedback mode(Generator only)
            enable     bool true or false
            """
            self.context._latent_feedback_toggle_action.setChecked(enable)


        def osc_set_gen_text(addr:str, text:str, encode:bool=True) -> None:
            """Set generator utterance text in gui textbox (Generator only)
            text        str new generate utterance text
            encode      bool if true, automatically encode the new text utterance
            """
            print(f"OSCfunc: set_gen_text {addr} | {text=} | {encode=}")
            # NOTE: One of the main gotcha's about Qt is that you cannot call any QWidget methods from any 
            # thread other than the main GUI thread. All of your communication must be done by emitting signals 
            # from the extra threads, which will forward to the main gui. So the following line will not work!
            # self.text_input.setText(text)
            signaller.set_gen_text.emit(text)

            if encode:
                print("ENCODE TEXT")
                self.context.btn_send_gen_text.click()

        def osc_set_alignment_as_token_idx(addr:str, tok_idx:int, force_paint:bool=False) -> None:
            """Set alignment of generator by token index (Generator Only)
            token_idx           int index of token
            force_paint         bool if true, toggles on attention painting
            """
            self.context.attention_graph.set_alignment_as_token_idx(tok_idx)
            print(f"Set alignment to token {tok_idx} - forced paint?: {force_paint}")

        def osc_set_alignment_normalized(addr:str, normalized_align:float, force_paint:bool=False) -> None:
            """Set alignment of generator by a normalized 0.0-1.0 value (Generator Only)
            normalized_align    float alignment value from 0-1 gets mapped to start-end token range
            force_paint         bool if true, toggles on attention painting
            """
            self.context.attention_graph.set_normalized_alignment(normalized_align)
            print(f"Set normalized alignment {normalized_align} - forced paint?: {force_paint}")

        def osc_set_temperature(addr:str, temp:float) -> None:
            """Set generator sampling temperature (Generator Only)
            temp    float temperature from 0.0-1.0 (can go up to 2.0 for more weird predictions)

            """
            self.context.set_temperature(temp)
            print(f"Set sampling temp {temp}")

        def osc_add_temperature(addr:str, temp:float) -> None:
            """Add a small value to the generator sampling temperature (Generator Only)
            temp    float value to add to temperature (usually below 1.0) - will clip at 0 and a max value

            """
            self.context.add_temperature(temp)
            print(f"Add {temp} to sampling temp")

        def osc_sampler_stop_at_end(addr:str, enable:bool) -> None:
            """ Enable/disable Sampler stop-at-end loop point
            enable     bool true or false
            """
            self.context._sampler_stop_at_end_toggle_action.setChecked(enable)

        def osc_set_sampler_step(addr:str, step:int, autoplay:bool=True):
            """Set sampler playhead absolute position (Sampler only)
            step        int absolute index of vocoder frame buffer to set sampler playhead, index wraps and can be negative
            autoplay    bool if true, sampler playback is triggered if it is not already playing
            """
            print(f"OSCfunc: set_sampler_step {addr} | {step=}")
            self.context.sampler_step(step=step, autoplay=autoplay)

        def osc_set_loop_text(addr:str, 
                text:str, n:int=-1, 
                start:bool=True, end:bool=True, reset:bool=True
                ) -> None:
            """Set sampler loop points to nth occurrence of matched text (Sampler only)
            text            str text to match in sampler history, can be a single token
            n               int which of the n text matches to loop, -1 is most recent, 0 oldest, etc..
            start           bool if true, updates the sampler loop start at the matched text
            end             bool if true, updates the sampler loop end at the matched text
            reset           bool if true resets sampler playhead to start of text match        
            """
            print(f"OSCfunc: set_loop_text {addr} | {text=} | {n=} | {start=} | {end=} | {reset=}")
            signaller.set_samp_text.emit(text)

            print("SET LOOP")
            self.context.loop_text(text=text, n=n, start=start, end=end, reset=reset)

        def osc_set_loop_index(addr:str, 
                start:int=None, end:int=None, 
                reset:bool=True,
                utterance:int=None,
                ) -> None:
            """Set sampler loop points by vocoder frame with option to index by utterance (Sampler only)
            start           int global start frame index
            end             int global end frame index
            reset           bool if true resets sampler playhead to start index
            utterance       global utterance index, if supplied changes start and end to be utterance-relative


            """
            print(f"OSCfunc: set_loop_index {addr} | {start=} | {end=} | {reset=} | {utterance=}")
            print("SET LOOP")
            self.context.loop_index(start=start, end=end, utterance=utterance, reset=reset)


        # END:: OSC Dispatcher Callback Functions -----------------------------------------------------------------------


        self.osc_dispatcher = pythonosc.dispatcher.Dispatcher()
        self.osc_dispatcher.set_default_handler(osc_unknown)
        
        # OSC Callback mappings
        self.osc_dispatcher.map("/generate", osc_generate) # TODO rename
        self.osc_dispatcher.map("/sampler", osc_sampler)
        self.osc_dispatcher.map("/pause", osc_pause)
        self.osc_dispatcher.map("/reset", osc_reset)

        self.osc_dispatcher.map("/set_bias", osc_set_bias)
        self.osc_dispatcher.map("/add_bias", osc_add_bias)
        self.osc_dispatcher.map("/reset_biases", osc_reset_biases)

        self.osc_dispatcher.map("/generate_stop_at_end", osc_generate_stop_at_end)
        self.osc_dispatcher.map("/alignment_mode", osc_alignment_mode)
        self.osc_dispatcher.map("/latent_feedback", osc_latent_feedback)
        self.osc_dispatcher.map("/set_gen_text", osc_set_gen_text)
        self.osc_dispatcher.map("/set_token", osc_set_alignment_as_token_idx)
        self.osc_dispatcher.map("/set_alignment", osc_set_alignment_normalized)
        self.osc_dispatcher.map("/set_temperature", osc_set_temperature)
        self.osc_dispatcher.map("/add_temperature", osc_add_temperature)

        self.osc_dispatcher.map("/sampler_stop_at_end", osc_sampler_stop_at_end)
        self.osc_dispatcher.map("/set_sampler_step", osc_set_sampler_step)
        self.osc_dispatcher.map("/set_loop_text", osc_set_loop_text)
        self.osc_dispatcher.map("/set_loop_index", osc_set_loop_index)



    def run(self):
        # Thread entry point
        print(f"Serving on {self.osc_server.server_address}")
        self.osc_server.serve_forever()
        print(f"Closing OSC Server...")

        # Need to update the GUI
        # signaller.set_text.emit("OSC Server Listening")


    def unserve_osc(self):
        if self.osc_server is not None:
            self.osc_server.shutdown()
            self.osc_server.server_close()
            self.osc_server = None
            print("Waiting for server to shutdown...")
            #self.osc_server_thread.join()
            print("Server Thread closed...")

    def serve_osc(self, address=("localhost", 7777)):

        if self.osc_server is not None:
            self.unserve_osc()

        self.osc_server = pythonosc.osc_server.ThreadingOSCUDPServer(address, self.osc_dispatcher)
        self.daemon = True
        #self.osc_server_thread = threading.Thread(target=run_osc_server, args=(self,), daemon=True)
        self.start()
        print(f"OSC Server Thread Started")


    # TODO can be improved to be co-generated together with dispatcher mappings
    def print_osc_api(self, basepath='/tungnaa'):
        instructions=""

        # TODO can be improved, this approach depends on undocumented internals of pythonosc.dispatcher
        if self.osc_dispatcher is not None:
            instructions += "OSC Control API:\n"
            for addr in self.osc_dispatcher._map.keys():
                handler = self.osc_dispatcher._map[addr][0] # take first available handler, although multiple are possible
                print(f"{addr} - {handler}")
                instructions += f"{addr} - {handler.callback.__doc__} \n" 

        instructions += f"""OSC Out API:
        {basepath}/latents
        {basepath}/status"""
        return instructions



class MainWindow(QtWidgets.QMainWindow):
    """
    Main Application Window
    """

    def __init__(self, 
        parent:QtWidgets.QWidget=None, 
        use_backend:bool=True, 
        backend:tungnaa.gui.backend.Backend=None, 
        sender:tungnaa.gui.senders.Sender=None,
        update_fps:int=100,
        osc_listen_addr:tuple[str,int]=('localhost', 1337),
        stress_gui=None,
        text=None,
        sampler_text=None,
        ):
        super().__init__(parent)
        self.version= "Alpha v0.1.0" # TODO: There's probably a tidier way to do this using Qt or python package management
        self.appname= f"T̴u̮n̵g̴na͠á {self.version}"

        self.stress_gui = stress_gui

        if sys.platform == 'darwin':
            self.setUnifiedTitleAndToolBarOnMac(True)

        self.mode = 'infer'
        
        self.update_fps = update_fps
        self.osc_listen_addr = osc_listen_addr
        self.use_backend = use_backend
        self.backend = backend
        self.sender = sender
        if self.backend is None:
            print("No backend provided, disabling backend")
            self.use_backend = False
        
        # OSC
        self.osc_controller = OSCController(context=self)
        
        # TODO: this line auto-runs the OSC server and creates a bug
        #     Currently needs to be activated from the gui, running this line starts the OSC server but does not reflect this status in the gui
        #self.serve_osc(tuple(osc_listen_addr))

        # Signals used by other threads to communicate with GUI objects defined in the main thread.

        # ---------------------------------------------------------------------
        # BUILD GUI
        # ---------------------------------------------------------------------
        #self.setWindowTitle(QtCore.QCoreApplication.applicationName())
        
        self.setWindowTitle(self.appname)
        self.main = QtWidgets.QWidget(self)
        # self.main.setStyleSheet("color: grey; background-color: black")
        self.setCentralWidget(self.main)
        self.settings_dialog = None # Settings modal dialogue
        self.app_icon = QtGui.QIcon()
        with resources.path("tungnaa.resources", "tungnaa_icon.png") as path:
            self.app_icon.addFile(path.as_posix())
        self.setWindowIcon(self.app_icon)
        self.setWindowIconText(self.appname)

        self.tungnaa_logo = QtWidgets.QLabel(parent=self.main)
        with resources.path("tungnaa.resources", "tungnaa_logo.png") as path:
            logo_pixmap= QtGui.QPixmap(path.as_posix())
        self.tungnaa_logo.setPixmap(logo_pixmap)
        self.tungnaa_logo.setMargin(1)
        self.tungnaa_logo.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignTop)
        self.tungnaa_version = QtWidgets.QLabel(parent=self.main)
        self.tungnaa_version.setText(self.version)
        version_font = QtGui.QFont()
        version_font.setPointSize(8)
        self.tungnaa_version.setFont(version_font)
        self.tungnaa_version.setMargin(0)
        self.tungnaa_version.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignTop)




        # Main App Toolbar
        self.toolbar_top = QtWidgets.QToolBar(self.main)
        self.toolbar_combined = QtWidgets.QWidget(self.main)
        self.toolbar_controls = QtWidgets.QToolBar(self.toolbar_combined)
        controls_layout = HBoxLayout()
        controls_layout.addWidget(self.toolbar_controls)
        controls_layout.addStretch(1)
        controls_layout.addLayout(VBoxLayout(self.tungnaa_logo, self.tungnaa_version))
        self.toolbar_combined.setLayout(controls_layout)
        self.addToolBar(self.toolbar_top) # Toolbar gets added to VLayout below

        self._generate_action = QtGui.QAction("Generate", self)
        self._generate_action.setStatusTip("Start autoregression")
        self._generate_action.triggered.connect(self.play_generate)

        self._sampler_action = QtGui.QAction("Sampler", self)
        self._sampler_action.setStatusTip("Sample generated latents")
        self._sampler_action.triggered.connect(self.play_sampler)

        self._pause_action = QtGui.QAction("Pause", self)
        self._pause_action.setStatusTip("Pause generation or sampler")
        self._pause_action.triggered.connect(self.pause)

        self._reset_action = QtGui.QAction("Reset", self)
        self._reset_action.setStatusTip("Reset autoregression history")
        self._reset_action.triggered.connect(self.reset_autoregression)

        self._alignment_paint_toggle_action = QtGui.QAction("Attention Painting", self)
        self._alignment_paint_toggle_action.setCheckable(True)
        self._alignment_paint_toggle_action.setStatusTip("Toggle Attention Painting")
        self._alignment_paint_toggle_action.toggled.connect(self.toggle_alignment_paint)

        self._latent_feedback_toggle_action = QtGui.QAction("Latent Feedback", self)
        self._latent_feedback_toggle_action.setCheckable(True)
        self._latent_feedback_toggle_action.setStatusTip("Feed latent manipulation back to model")
        self._latent_feedback_toggle_action.toggled.connect(self.toggle_latent_feedback)

        self._generate_stop_at_end_toggle_action = QtGui.QAction("Generator Stop At End", self)
        self._generate_stop_at_end_toggle_action.setCheckable(True)
        self._generate_stop_at_end_toggle_action.setStatusTip("Pause automatically at end of utterance text in generate mode")
        self._generate_stop_at_end_toggle_action.toggled.connect(self.toggle_generate_stop_at_end)

        self._sampler_stop_at_end_toggle_action = QtGui.QAction("Sampler Stop At End", self)
        self._sampler_stop_at_end_toggle_action.setCheckable(True)
        self._sampler_stop_at_end_toggle_action.setStatusTip("Pause automatically at loop end point in sampler mode")
        self._sampler_stop_at_end_toggle_action.toggled.connect(self.toggle_sampler_stop_at_end)

        self._latent_bias_reset_action = QtGui.QAction("Bias Reset", self)
        self._latent_bias_reset_action.setStatusTip("Reset all latent biases to 0.0")
        self._latent_bias_reset_action.triggered.connect(self._reset_latent_biases)

        self._settings_action = QtGui.QAction("Settings", self)
        self._settings_action.setStatusTip("Adjust OSC, MIDI and Audio settings")
        self._settings_action.triggered.connect(self.open_settings)
        self.temperature_slider = DoubleSlider(orientation=QtCore.Qt.Horizontal, decimals=3, parent=self)
        self.temperature_slider.setMaximum(2.0)
        self.temperature_slider.setMinimum(0.0)
        self.temperature_slider.setValue(1)
        self.temperature_slider.doubleValueChanged.connect(lambda val: self._temperature_adjust(val))
        self.temperature_slider.setStatusTip("sampling temperature")


        self.toolbar_top.addAction(self._alignment_paint_toggle_action)
        self.toolbar_top.addAction(self._latent_feedback_toggle_action)
        self.toolbar_top.addAction(self._sampler_stop_at_end_toggle_action)
        self.toolbar_top.addAction(self._generate_stop_at_end_toggle_action)
        self.toolbar_top.addAction(self._settings_action)

        self.toolbar_controls.addAction(self._generate_action)
        self.toolbar_controls.addAction(self._sampler_action)
        self.toolbar_controls.addAction(self._pause_action)
        self.toolbar_controls.addAction(self._reset_action)
        self.toolbar_controls.addAction(self._latent_bias_reset_action)
        


        # Tabbed Interface for Generator/Sampler
        self.tabwidget = QtWidgets.QTabWidget(self.main);

        # Generator Tab
        self.generator_widget = QtWidgets.QWidget(self.tabwidget);

        # See https://doc.qt.io/qt-6/qtextedit.html for all the fun slots for textedit
        self.gen_text_input = QtWidgets.QTextEdit(parent=self.generator_widget)
        self.gen_text_input.setMaximumHeight(100)
        self.gen_text_input.setFontPointSize(24)
        self.gen_text_input.setPlainText(text or "<It took me a while to have a voice. And now that I have one, I am not going to be silent.>")
        self.gen_text_input.setStatusTip("enter text to generate new utterances")
        signaller.set_gen_text.connect(self.gen_text_input.setPlainText)
        #self.text_input.setPlainText("It took me a while to find a voice, and now that I have one, I am not going to be silent.") # todo - add a "random phrase" generator?
        self.btn_send_gen_text = QtWidgets.QPushButton("Encode", parent=self.generator_widget)
        self.btn_send_gen_text.clicked.connect(self.encode_text)
        self.btn_send_gen_text.setStatusTip("send text to model")
        self.text_encoding_status = QtWidgets.QLabel(parent=self.generator_widget)
        self.text_encoding_status.setText("ʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭ encoder feedback... ʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭ")


        self.attention_graph = AlignmentGraph(parent=self.generator_widget, image_clicked=self.image_clicked)
        self.attention_graph.setStatusTip("text-audio alignments (click to jump)")
        
        self._generator_layout = VBoxLayout(
                self.gen_text_input, 
                self.btn_send_gen_text,
                self.text_encoding_status,
                self.attention_graph,
                spacing=0, margins=(0,0,0,0)
        )
        self.generator_widget.setLayout(self._generator_layout);


        # Sampler Tab
        self.sampler_widget = QtWidgets.QWidget(self.tabwidget);

        self.samp_text_input = QtWidgets.QTextEdit(self.sampler_widget)
        self.samp_text_input.setMaximumHeight(100)
        self.samp_text_input.setFontPointSize(24)
        self.samp_text_input.setPlainText(sampler_text or "took")
        self.samp_text_input.setStatusTip("enter text to search through previous utterances in sampler mode")
        signaller.set_samp_text.connect(self.samp_text_input.setPlainText)
        self.btn_send_samp_text = QtWidgets.QPushButton("Sample", parent=self.sampler_widget)
        self.btn_send_samp_text.clicked.connect(self.loop_text)
        self.btn_send_samp_text.setStatusTip("set sampler loop points by text")


        self._sampler_layout = VBoxLayout(
            VBoxLayout(
                self.samp_text_input, 
                self.btn_send_samp_text,
                spacing=0, margins=(0,0,0,0)
            ),
            spacing=0, margins=(12,0,12,0),
        )
        self.sampler_widget.setLayout(self._sampler_layout)

        self.tabwidget.addTab(self.generator_widget, "Generator")
        self.tabwidget.addTab(self.sampler_widget, "Sampler")        



        # StatusBar
        self.statusbar = QtWidgets.QStatusBar(self.main)
        self.setStatusBar(self.statusbar)

        self.latents = RaveLatents(self.main)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonFollowStyle)


        self._main_layout = VBoxLayout(
            self.toolbar_combined,
            self.tabwidget,
            self.latents,
            self.temperature_slider,
            spacing=0, margins=(0,0,0,0)
        )
        self.main.setLayout(self._main_layout)

        # Keyboard Shortcuts (TODO: put these in a menu somewhere)
        self._encode_action = QtGui.QAction("Encode", self)
        #self._encode_action.autoRepeat = False
        self._encode_action.triggered.connect(self.btn_send_gen_text.click)
        self._encode_action.setShortcut("Ctrl+Return") # TODO: Shift+Return overridden by QTextEdit ... need to subclass?
        self.addAction(self._encode_action) # add to the main window as a global shortcut

        self.gui_update_timer = QtCore.QTimer(self)
        self.gui_update_timer.timeout.connect(self.update)
        
        # Used to generate fake data in update() when no backend is provided
        self.frame = 0
        self.tok = 30
        self.max_tok = 40
        self.gui_update_timer.start((1.0 / self.update_fps) * 1000)

        self._generate_stop_at_end_toggle_action.setChecked(False) # make the default to babble like a river
        self._sampler_stop_at_end_toggle_action.setChecked(True)


    def image_clicked(self, ev):
        # print(ev.pos)
        self.backend.set_momentary_alignment((ev.pos().x(), 1))
        self.backend.set_state_by_step(int(ev.pos().y()))
        self.backend.generate()
        # self.backend.set_momentary_alignment(
        #     self.attention_graph.get_slidervalue_as_params())

    def update(self):
        """
        update method runs on a timer
        empties the queue from the backend and updates gui elements
        """
        self.frame += 1
        new_data = list()
        # empty message queue, update attention graph & RAVE latents
        if self.use_backend: 

            if self.stress_gui is not None:
                t = time.time_ns()
                while time.time_ns() - t < self.stress_gui*1e9:
                    pass

            try:
                if self.latents.is_init:
                    self.backend.set_biases(self.latents.get_biases())
                if self.mode == 'paint':
                    self.backend.set_alignment(
                        self.attention_graph.get_slidervalue_as_params())
                elif self.mode == 'infer':
                    self.backend.set_alignment(None)
                else:
                    raise ValueError(f"Unknown attention mode: {self.mode} - must be <infer|paint>")

                # for _ in range(self.backend.frontend_q.qsize()):
                while self.backend.parent_conn.poll():
                    try:
                        # framedict = self.backend.frontend_q.get_nowait()
                        framedict = self.backend.parent_conn.recv()
                        if not self.latents.is_init and 'num_latents' in framedict and 'use_pitch' in framedict:
                            self.latents._init(
                                framedict['num_latents'], framedict['use_pitch'])

                        new_data.append(framedict)
                    except queue.Empty as ex:
                        break
            except Exception as e:
                print(e)
                exit(0) ### debug
            
        else: # Do not use backend, instead generate random data.. useful for testing the gui (maybe?)
            new_attn_frame = np.zeros((1,self.attention_graph.num_encodings), dtype=np.float32)        
            new_latent_frame = np.random.rand(1,8) * 0.5
            new_latent_frame = new_latent_frame + self.latents.get_biases()
            if self.mode == 'infer':
                if random.random() < 0.2:
                    self.tok += random.choice([-1, 1])
            elif self.mode == 'paint':
                self.tok = self.attention_graph.alignment_slider.value()            
            else:
                raise ValueError(f"Unknown attention mode: {self.mode} - must be <infer|paint>")

            if self.tok >= self.max_tok-1:
                self.tok = self.max_tok - 1
            else:
                new_attn_frame[0,self.tok + 1] = 0.5
            if self.tok <= 0:
                self.tok = 0
            else:
                new_attn_frame[0,self.tok - 1] = 0.5
            new_attn_frame[0,self.tok] = 1.0

            new_data.append({
                'latent_t': new_latent_frame, 'align_t': new_attn_frame})
            
        # iterate through new_data and update the gui
        for datadict in new_data:
            if datadict.get('reset', False):
                # self.finish_reset(num_tokens=datadict['align_t'].shape[-1])
                self.finish_reset(text=datadict['text'])
            if 'align_t' in datadict and not datadict.get('sampler', False):
                self.attention_graph.addFrame(
                    datadict['align_t'], datadict['text'])
        
        latents = [d['latent_t'] for d in new_data if 'latent_t' in d]
        if len(latents) and self.latents.is_init:
            self.latents.set_latents(values=latents[-1])

    def finish_reset(self, text):
        """
        Called when backend signals a reset.
        """
        self.text_encoding_status.setText(f"ʭʬʭʬʭʬʭʬ encoded {len(text)} embeddings ʬʭʬʭʬʭʬʭ")
        self.attention_graph.reset(text=text)

    def closeEvent(self, e:QtCore.QEvent):
        """
        Cleanup
        """
        # TODO: cleanup OSC/networking connections
        print(f"Application Close {e}")
        self.backend.cleanup()
        e.accept()
        #e.ignore() # Under some conditions ignore app close?

    def _temperature_adjust(self, temp=1.0) -> None:
        """
        Private method adjust model step inference temperature
        """
        if self.use_backend:
            self.set_temperature(temp)

    def _reset_latent_biases(self):
        """Private method to reset latent biases to 0.0"""
        if self.latents.is_init:
            self.latents.reset_latent_biases(0.0)

    def encode_text(self):
        """
        Send input text to the text encoder backend
        """
        if self.use_backend:
            txtval = self.gen_text_input.toPlainText()
            print(f"Encoding: {txtval}")
            self.text_encoding_status.setText("ʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭ encoding.... ʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭʬʭ")
            # this returns the text with start/end tokens added and loop points stripped
            self.backend.set_text(text=txtval)
        else:
            print("No backend enabled to encode text: ignoring...")

    def sampler_step(self, step:int, autoplay:bool) -> None:
        self.backend.set_sampler_step(step)
        if autoplay:
            # TODO: Once sampler mode is a toggle rather than a trigger, need to implement something like this to put the GUI in sampler mode...
            # if gui not_in_sampler_mode                
            #     self._sampler_action.toggle(True)
            # else:
            self.backend.sampler()

    def loop_text(self, **kw):
        """
        Send input text to the sampler backend
        """
        if not self.use_backend:
            print("No backend enabled to loop text: ignoring...")
            return
        if 'text' not in kw:
            kw['text'] = self.samp_text_input.toPlainText()
        self.backend.set_sampler_loop_text(**kw)

    def loop_index(self, **kw):
        """
        Send input text to the sampler backend
        """
        if not self.use_backend:
            print("No backend enabled to loop indices: ignoring...")
            return
        self.backend.set_sampler_loop_index(**kw)

    def play_generate(self, val:bool):
        print(f"Play autoregressive frame generator")
        ### TODO: replace this
        # if self.backend.text_rep is None:
            # self.encode_text()
        if self.attention_graph.text is None:
            self.encode_text()
        self.backend.generate()

    def play_sampler(self, val:bool):
        print(f"Play sampler")
        self.backend.sampler()

    def pause(self, val:bool):
        print(f"Pause generation or sampler")
        self.backend.pause()

    def reset_autoregression(self, val:bool):
        print(f"Reset autoregression history")
        # self.attention_graph.reset()
        self.backend.reset()

    def set_temperature(self, temp:float) -> None:
        """
        Set inference temperature (used by OSC/MIDI)

        Args:
            temp  inference temperature from 0.0-2.0
        """
        if temp > 2.0:
            temp=2.0
        elif temp < 0:
            temp=0
        self.temperature_slider.setValue(temp)
        self.backend.set_temperature(temp)

    def add_temperature(self, temp:float) -> None:
        """
        Add a small value to the inference temperature (used by OSC/MIDI)

        Args:
            temp    value that will be added to inference temperature, usually small, temperature will clip at [0,2]
        """
        self.set_temperature(self.temperature_slider.value() + temp)

    def toggle_alignment_paint(self, toggle:bool):
        if toggle:
            self.mode = 'paint'
        else:
            self.mode = 'infer'
        print(f"Alignment Mode Changed To:{self.mode}")

    def toggle_latent_feedback(self, toggle:bool):
        self.backend.set_latent_feedback(toggle)
        print(f"Latent Feedback status changed to:{toggle}")

    def toggle_generate_stop_at_end(self, toggle:bool):
        self.backend.set_generate_stop_at_end(toggle)
        print(f"Stop Generate At End status changed to:{toggle}")

    def toggle_sampler_stop_at_end(self, toggle:bool):
        self.backend.set_sampler_stop_at_end(toggle)
        print(f"Stop Sampler At End status changed to:{toggle}")

    def set_alignment_mode(self, mode:str='infer') -> None:
        """
        Set alignment mode directly (used by OSC/MIDI)
        """
        if mode in ['infer', 'paint']:
            self.mode = mode
            self._alignment_paint_toggle_action.setChecked((mode == 'paint'))
            print(f"Alignment Mode changed to: {mode}")

    def open_settings(self):
        if self.settings_dialog is None:
            self.settings_dialog = SettingsDialog(context=self)
        self.settings_dialog.show() # use show() to display a modeless dialog

    def serve_osc(self, address=("localhost", 7777)):
        self.osc_controller.unserve_osc()
        self.osc_controller.serve_osc(address=address)

    def unserve_osc(self):
        self.osc_controller.unserve_osc()

# Main entry point for Fire
def main(
    # models
    tts:Path='tungnaa_119_vctk',
    vocoder:Path=None,
    repo='Intelligent-Instruments-Lab/tungnaa-models-public',
    # output modes
    synth_audio:bool=True,
    latent_audio:bool=False,
    latent_osc:bool=False,
    # audio driver
    audio_out:typing.Union[str,int]=None,
    audio_block:int=2048,
    # OSC
    osc_out_addr:str='localhost:57120',
    osc_out_path:str='/tungnaa',
    osc_out_scsynth:bool=False,
    osc_in_addr:str='localhost:1337',
    # misc
    no_backend:bool=False,
    buffer_frames:int=1,
    stress_gui:float=None,
    jit:bool=False,
    profile:bool=False,
    text:str|None=None,
    sampler_text:str|None=None,
):
    """Tungnaa Text to Voice
    
    Args:
        tts: path to TTS model, or name of model in repo
        vocoder: path to vocoder model
        repo: huggingface repo to search for models

        synth_audio: send stereo audio from Python
        latent_audio: pack latents into a mono audio signal, 
            to be unpacked into multiple signals and decoded elsewhere
        latent_osc: send latents over OSC, 
            to be converted to audio signals and decoded elsewhere

        audio_out: 'default', or audio output device number (see 'tungnaa device' for a list of devices)
        audio_block: audio block size in samples

        osc_out_addr: host:port to send OSC data to. 
            By default sends to localhost:57120
        osc_out_path: The OSC path prefix of messages sent by this app. 
            Uses /tungnaa by default.
        osc_out_scsynth: format OSC for sending directly to scsynth
        osc_in_addr: host:port for recieving OSC control messages from clients.   
            By default listens on localhost:1337
        buffer_frames: compute ahead this many model steps
            increases latency, but may mitigate dropouts
        stress_gui: for testing, add this many seconds delay in `update`
        
    """

    if audio_out == 'default':
        audio_out = None
    if audio_out is not None:
        if isinstance(audio_out, str) and audio_out.isdecimal():
            audio_out = int(audio_out)
        elif not isinstance(audio_out, int):
            raise TypeError(f"Invalid audio device id '{audio_out}', must be either an integer or 'default'")

    app = QtWidgets.QApplication(sys.argv)
    QtCore.QCoreApplication.setOrganizationName("Intelligent Instruments Lab")
    QtCore.QCoreApplication.setApplicationName("Tungnaa")
    QtCore.QCoreApplication.setApplicationVersion(QtCore.qVersion())

    if latent_osc:
        if osc_out_scsynth: # Send latents to scsynth via OSC using default scsynth address localhost:57110
            sender = tungnaa.gui.senders.SCSynthDirectOSCSender(
                host='127.0.0.1',
                port=57110,
                bus_index=64,
                latency=0.2
            )
        else: # Send latents out via Generic OSC
            osc_host, osc_port = osc_out_addr.split(':')
            osc_port = int(osc_port)
            sender = tungnaa.gui.senders.GenericOSCSender(
                host=osc_host, 
                port=osc_port, 
                lroute=f'{osc_out_path}/latents',
                sroute=f'{osc_out_path}/status'
            )
    else:
        sender = None

    # if tts is not a local file, download model from repo
    # also sets the vocoder unless it it explicitly set to something else
    try:
        with open(tts): pass
    except FileNotFoundError:
        print(f'searching remote repo for model "{tts}"...')
        tts, vocoder = dl_model(repo, tts, vocoder)

    backend = tungnaa.gui.backend.Backend(
        checkpoint=tts, 
        rave_path=vocoder,
        audio_out=audio_out,
        audio_block=audio_block,
        # audio_channels=audio_channels,
        synth_audio=synth_audio,
        latent_audio=latent_audio,
        latent_osc=latent_osc,
        osc_sender=sender,
        buffer_frames=buffer_frames,
        jit=jit,
        profile=profile,
        )
    
    osc_host, osc_port = osc_in_addr.split(':')
    osc_port = int(osc_port)
    win = MainWindow(
        use_backend=(not no_backend), 
        sender=sender,
        backend=Proxy(backend),
        osc_listen_addr=(osc_host, osc_port),
        stress_gui=stress_gui,
        text=text,
        sampler_text=sampler_text,
        )
    
    available_geometry = win.screen().availableGeometry()
    win.resize(available_geometry.width() / 2 * 1.063, available_geometry.height())
    win.move((available_geometry.width() - win.width()), 0)
    win.show()

    # backend.start_stream()

    sys.exit(app.exec())

if __name__ == "__main__":
    fire.Fire(main)



from typing import Optional, Callable, Union, Tuple, List
from numbers import Number
from enum import Enum, Flag
import re
from collections import defaultdict
import cProfile, pstats, io

import threading, os
from threading import Thread, RLock
from queue import Queue
import time
from contextlib import contextmanager
# from fractions import Fraction

import fire

# import numpy as np
# import numpy.typing as npt
import torch
from torch import Tensor
torch.set_num_threads(1)

import sounddevice as sd

from tungnaa import TacotronDecoder, lev

OutMode = Flag('OutMode', ['SYNTH_AUDIO', 'LATENT_AUDIO', 'LATENT_OSC'])
StepMode = Enum('StepMode', ['PAUSE', 'SAMPLER', 'GENERATION'])

def print_audio_devices():
    devicelist = sd.query_devices()
    for dev in devicelist:
        print(f"{dev['index']}: '{dev['name']}' {dev['hostapi']} (I/O {dev['max_input_channels']}/{dev['max_input_channels']}) (SR: {dev['default_samplerate']})")
    return devicelist

class Profiler:
    def __init__(self, enabled):
        self.enabled = enabled
        self.stack = []
        self.mean_wall_ms = defaultdict(float)
        self.mean_proc_ms = defaultdict(float)
        self.n = defaultdict(int)
    @contextmanager
    def __call__(self, label, detail=False):
        if self.enabled:
            if detail:
                pr = cProfile.Profile()
                pr.enable()
                # pr = torch.profiler.profile()
                # pr.__enter__()
            self.stack.append((time.time_ns(), time.process_time_ns()))
        yield None
        if self.enabled:
            t_wall, t_proc = self.stack.pop()
            if detail:
                # pr.__exit__(None, None, None)
                pr.disable()
            proc_ms = (time.process_time_ns() - t_proc)*1e-6
            wall_ms = (time.time_ns() - t_wall)*1e-6
            self.n[label] += 1
            self.mean_wall_ms[label] = (self.mean_wall_ms[label]*(self.n[label]-1) + wall_ms) / self.n[label]
            self.mean_proc_ms[label] = (self.mean_proc_ms[label]*(self.n[label]-1) + proc_ms) / self.n[label]
            # print(
            #     'average', ' '*len(self.stack), label, 
            #     f'wall: {int(self.mean_wall_ms[label])} ms,',
            #     f'cpu: {int(self.mean_proc_ms[label])} ms')
            if self.n[label] > 10 and wall_ms > 2*self.mean_wall_ms[label]:
                print(
                    'slower than 2x average:',
                    '  '*len(self.stack), 
                    label, 
                    f'wall: {int(wall_ms)} ms', 
                    f'cpu: {int(proc_ms)} ms', 
                    f'(average wall {int(self.mean_wall_ms[label])} ms, cpu {int(self.mean_proc_ms[label])} ms)')
                if detail:
                    # print(pr.key_averages().table(
                        # sort_by="self_cpu_time_total", row_limit=4))
                    s = io.StringIO()
                    ps = pstats.Stats(pr, stream=s)
                    ps.sort_stats('tottime')
                    ps.print_stats(4)
                    print(s.getvalue())

class Utterance(list):
    def __init__(self, text):
        self.text = text


class Backend:
    def __init__(self,
        checkpoint:str,
        rave_path:str|None=None,
        audio_in:str|None=None,
        audio_out:str|int=None,
        audio_block:int|None=None,
        audio_channels:int|None=None,
        # sample_rate:int=None,
        synth_audio:bool|None=None,
        latent_audio:bool|None=None,
        latent_osc:bool|None=None,
        osc_sender=None,
        buffer_frames:int=1,
        profile:bool=True,
        jit:bool=False,
        max_model_state_storage:int=512,
        ):
        """
        Args:
            checkpoint: path to tungnaa checkpoint file
            rave_path: path to rave vocoder to use python sound
            audio_in: sounddevice input name/index if using python audio
            audio_out: sounddevice output name/index if using python audio
            audio_block: block size if using python audio
            synth_audio: if True, vocode in python and send stereo audio
            latent_audio: if True, pack latents into a mono audio channel
            latent_osc: if True, send latents over OSC
            buffer_frames: process ahead this many model steps
            profile: if True, print performance profiling info
            jit: if True, compile tungnaa model with torchscript
        """
        self.jit = jit
        self.profile = Profiler(profile)
        # default output modes
        if synth_audio is None:
            synth_audio = rave_path is not None
        if latent_audio is None:
            latent_audio = rave_path is None
        if latent_osc is None:
            latent_osc = False

        self.reset_values = {}

        out_mode = OutMode(0)
        if synth_audio: out_mode |= OutMode.SYNTH_AUDIO
        if latent_audio: out_mode |= OutMode.LATENT_AUDIO
        if latent_osc: out_mode |= OutMode.LATENT_OSC
        self.out_mode = out_mode
        print(f'{self.out_mode=}')

        if audio_channels is None:
            audio_channels = 0
            if OutMode.SYNTH_AUDIO in self.out_mode:
                self.synth_channels = (0,1)
                audio_channels += 2
            else:
                self.synth_channels = tuple()
            if OutMode.LATENT_AUDIO in self.out_mode:
                self.latent_channels = (audio_channels,)
                audio_channels += 1
            else:
                self.latent_channels = tuple()
        else:
            raise NotImplementedError(
                "setting audio_channels not currently supported")
        print(f'{audio_channels=}')

        self.step_mode = StepMode.PAUSE
        self.generate_stop_at_end = False
        self.sampler_stop_at_end = False

        self.latent_biases = []

        # self.temperature = 0.5
        self.temperature = 1.

        # generation
        self.gen_loop_start = 0
        self.gen_loop_end = None
        # sampler
        self.sampler_loop_start = 0
        self.sampler_loop_end = None
        # self.sampler_utterance = -1
        self.sampler_step = 0

        self.osc_sender = osc_sender

        self.frontend_conn = None

        self.max_model_state_storage = max_model_state_storage

        ### move heavy init out of __init__, so it only runs in child process
        ### (Backend.run is called by Proxy._run)
        self.init_args = (buffer_frames, audio_out, checkpoint, rave_path, audio_block, audio_channels)

        # call this from both __init__ and run
        # torch.multiprocessing.set_sharing_strategy('file_system')


    def run(self, conn):
        """run method expected by Proxy"""
        # call this from both __init__ and run
        # torch.multiprocessing.set_sharing_strategy('file_system')

        self.frontend_conn = conn
        # print(f'{self.frontend_conn=} {threading.get_native_id()=} {os.getpid()=}')

        buffer_frames, audio_out, checkpoint, rave_path, audio_block, audio_channels = self.init_args

        self.load_tts_model(checkpoint=checkpoint)

        ### synthesis in python:
        # load RAVE model
        self.load_vocoder_model(rave_path=rave_path)

        ### audio output:
        # make sounddevice stream
        if self.out_mode:
            devicelist = sd.query_devices()
            
            # None throws an error on linux, on mac it uses the default device
            # Let's make this behavior explicit.
            if audio_out is None:
                audio_out = sd.default.device[1]

            audio_device = None
            for dev in devicelist:
                if audio_out in [dev['index'], dev['name']]:
                    audio_device = dev
                    # audio_device_sr = dev['default_samplerate']
                print(f"{dev['index']}: '{dev['name']}' {dev['hostapi']} (I/O {dev['max_input_channels']}/{dev['max_input_channels']}) (SR: {dev['default_samplerate']})")

            # this should not be an error
            # None uses the default device on macOS
            # if audio_device is None:
                # raise RuntimeError(f"Audio device '{audio_out}' does not exist.")

            print(f"USING AUDIO OUTPUT DEVICE {audio_out}:{audio_device}")

            self.active_frame:torch.Tensor = None 
            self.future_frame:Thread = None
            self.frame_counter:int = 0

            sd.default.device = audio_out
            if self.rave_sr:
                if audio_device and audio_device['default_samplerate'] != self.rave_sr:
                    # this should not be an error. On OSX/CoreAudio you can set the device sample rate to the model sample rate.
                    # however on Linux/JACK this throws a fatal error and stops program execution, requiring you to restart Jack to change the sampling rate

                    # TODO: also check RAVE block size vs. audio device block size if possible
                    print("\n------------------------------------");
                    print(f"WARNING: Device default sample rate ({audio_device['default_samplerate']}) and RAVE model sample rate ({self.rave_sr}) mismatch! You may need to change device sample rate manually on some platforms.")
                    print("------------------------------------\n");

                sd.default.samplerate = self.rave_sr # could cause an error if device uses a different sr from model
                print(f"RAVE SAMPLING RATE: {self.rave_sr}")
                print(f"DEVICE SAMPLING RATE: {audio_device['default_samplerate']}")

            # TODO: Tungnaa only uses audio output. Shouldn't we always be using sd.OutputStream?
            try:
                assert len(audio_out)==2
                stream_cls = sd.Stream
            except Exception:
                stream_cls = sd.OutputStream

            self.stream = stream_cls(
                callback=self.audio_callback,
                samplerate=self.rave_sr, 
                blocksize=audio_block, 
                #device=(audio_in, audio_out)
                device=audio_out,
                channels=audio_channels
            )
            
            if self.rave_sr:
                assert self.stream.samplerate == self.rave_sr, f"""
                failed to set sample rate to {self.rave_sr} from sounddevice
                """
        else:
            self.stream = None

        self.text = None
        self.text_rep = None
        self.align_params = None
        self.momentary_align_params = None
        self.step_to_set = None

        self.one_shot_paint = False
        self.latent_feedback = False

        self.lock = RLock()
        self.text_update_thread = None

        # self.frontend_q = Queue()
        self.audio_q = Queue(buffer_frames)     
        self.trigger_q = Queue(buffer_frames)     

        self.states = []

        self.needs_reset = False

        self.step_thread = Thread(target=self.step_loop, daemon=True)
        self.step_thread.start()
 
    def step_loop(self):
        """model stepping thread
        
        for each timestamp received in trigger_q
        send audio frames in audio_q
        """
        while True:
            t = self.trigger_q.get()
            with self.profile('step'):
                frame = self.step(t)
            if frame is not None:
                # with self.profile('frame.numpy'):
                frame = frame.numpy()
            self.audio_q.put(frame)
    
    def load_tts_model(self, checkpoint):
        """helper for loading TTS model, called on initialization but can also be called from GUI"""
        self.model = TacotronDecoder.from_checkpoint(checkpoint)

        # not scripting text encoder for now
        # if self.model.text_encoder is None:
            # self.text_model = TextEncoder()
        # else:
        self.text_model = self.model.text_encoder
        self.model.text_encoder = None

        # def _debug(m):
        #     # print({(k,type(v)) for k,v in m.__dict__.items()})
        #     for k,v in m.__dict__.items():
        #         # print('ATTR', k)
        #         if 'tensor' in str(type(v)).lower(): 
        #             print('TENSOR', k)
        #     for m_ in m.modules():
        #         if m_ != m:
        #             # print('MODULE', m_)
        #             _debug(m_)
        # _debug(self.model)

        if self.jit:
            for m in self.model.modules():
                if hasattr(m, 'parametrizations'):
                    torch.nn.utils.parametrize.remove_parametrizations(
                        m,'weight')
            self.model = torch.jit.script(self.model)
        self.model.eval()
        # self.model.train()

        print(f'{self.num_latents=}')

        # print(f'{self.model.frame_channels=}')
        self.use_pitch = hasattr(self.model, 'pitch_xform') and self.model.pitch_xform

    def load_vocoder_model(self, rave_path):
        """helper for loading RAVE vocoder model, called on initialization but can also be called from GUI"""
        # TODO: The audio engine sampling rate gets set depending on the sampling rate from the vocoder. 
        #   When loading a new vocoder model we need to add some logic to make sure the new vocoder has the same sample rate as the audio system.
        if OutMode.SYNTH_AUDIO in self.out_mode:
            assert rave_path is not None
            self.rave = torch.jit.load(rave_path, map_location='cpu')
            self.rave.eval()
            self.block_size = int(self.rave.decode_params[1])
            try:
                self.rave_sr = int(self.rave.sampling_rate)
            except Exception:
                self.rave_sr = int(self.rave.sr)
            with torch.inference_mode():
                # warmup
                if hasattr(self.rave, 'full_latent_size'):
                    latent_size = self.rave.latent_size + int(
                        hasattr(self.rave, 'pitch_encoder'))
                else:
                    latent_size = self.rave.cropped_latent_size
                self.rave.decode(torch.zeros(1,latent_size,1))
        else:
            self.rave = None
            self.rave_sr = None
    
    @property
    def num_latents(self):
        return self.model.frame_channels
    
    def start_stream(self):
        """helper for start/sampler"""
        # if self.out_mode in (OutMode.LATENT_AUDIO, OutMode.LATENT_OSC):
        #     if not self.loop_thread.is_alive():
        #         self.run_thread = True
        #         self.loop_thread.start()
        # if self.out_mode in (OutMode.LATENT_AUDIO, OutMode.SYNTH_AUDIO):
        # if self.out_mode is not None:
        if not self.stream.active:
            self.stream.start()

    def generate(self):
        """
        start autoregressive alignment & latent frame generation
        """
        # if self.step_mode != StepMode.GENERATION:
            # self.needs_reset = True
        self.step_mode = StepMode.GENERATION
        self.start_stream()

    def pause(self):
        """
        pause generation or sampler
        """
        self.step_mode = StepMode.PAUSE

    def sampler(self):
        """
        start sampler mode
        """
        self.step_mode = StepMode.SAMPLER
        self.start_stream()

    def reset(self):
        """
        reset the model state and alignments history
        """
        self.needs_reset = True

    def cleanup(self):
        """
        Cleanup any resources
        """
        self.step_mode = StepMode.PAUSE
        self.run_thread = False
        # should probably do this more gracefully
        exit(0) # exit the backend process

    def set_text(self, text:str) -> int:
        """
        Compute embeddings for & store a new text, replacing the old text.
        Returns the number of embedding tokens

        Args:
            text: input text as a string

        Returns:
            length of text in tokens
        """
        if (
            self.text_update_thread is not None 
            and self.text_update_thread.is_alive()
        ):
            print('warning: text update still pending')

        # TODO: more general text preprocessing
        text, start, end = self.extract_loop_points(text)
        tokens, text, idx_map = self.text_model.tokenize(text)
        # print(start, end, idx_map)
        # NOTE: positions in text may change fron tokenization (end tokens)
        if start < len(idx_map):
            start = idx_map[start]
        else:
            start = 0
        if end is not None and end < len(idx_map):
            end = idx_map[end]
        else:
            end = None

        # text processing runs in its own thread
        self.text_update_thread = Thread(
            target=self._update_text, 
            args=(text, tokens, start, end), daemon=True)
        self.text_update_thread.start()

        return text

    def extract_loop_points(self, text, tokens='<>'):
        """helper for `set_text`"""
        start_tok, end_tok = tokens
        # TODO: could look for matched brackets, have multiple loops...
        start = text.find(start_tok) # -1 if not found
        # if not len(text):
        #     return text, None, None
        if start < 0:
            start = 0
        else:
            text = text[:start]+text[start+1:]
        end = text.find(end_tok) # -1 if not found
        if end < 0: 
            end = None
        else:
            text = text[:end]+text[end+1:]
            end = max(0, end - 1)
        # print(text, end)
        return text, start, end

    def _update_text(self, text, tokens, start, end):
        """runs in a thread"""
        # store the length of the common prefix between old/new text
        # if self.text is None:
        #     self.prefix_len = 0
        # else:
        #     for i,(a,b) in enumerate(zip(text, self.text)):
        #         print(i, a, b)
        #         if a!=b: break
        #     self.prefix_len = i

        # store a mapping from old text positions to new
        if self.text is None:
            text_index_map = lambda x:x
        else:
            _,_,tm = lev(self.text, text)
            text_index_map = lambda x: tm[x] if x < len(tm) else len(text)-1

        # lock should keep multiple threads from trying to run the text encoder
        # at once in case `input_text` is called rapidly
        with self.lock:
            with torch.inference_mode():
                self.reset_values = dict(
                    text_rep = self.text_model.encode(tokens),
                    text_index_map = text_index_map,
                    text_tokens = tokens,
                    text = text,
                    gen_loop_start = start,
                    gen_loop_end = end,
                )
            self.reset()

    def set_biases(self, biases:List[float]):
        self.latent_biases = biases

    def set_alignment(self, 
            align_params:Optional[Tuple[float, float]]
        ) -> None:
        """
        Send alignment parameters to the backend. If None is passed, alignment painting is off.

        Args:
            align_params: [loc, scale] or None
        """
        self.align_params = align_params
    
    def set_momentary_alignment(self,             
            align_params:Optional[Tuple[float, float]]
        ) -> None:
        """
        Alignment will be set on the next frame only
        """
        self.momentary_align_params = align_params

    def set_state_by_step(self, step):
        self.step_to_set = step

    def set_latent_feedback(self, b:bool):
       self.latent_feedback = b

    def set_generate_stop_at_end(self, b:bool):
        self.generate_stop_at_end = b

    def set_sampler_stop_at_end(self, b:bool):
        self.sampler_stop_at_end = b

    def set_sampler_step(self, step:int):
        self.sampler_step = step % self.total_steps()

    def set_sampler_loop_index(self, 
            start:int=None, end:int=None, 
            utterance:int=None, 
            reset:bool=True):
        """
        Args:
            start: loop start step
            end: loop end step
            utterance: sampler utterance
            reset: if True, immediately go to loop start
        """
        if len(self.states)==0: return

        def wrap(x, n, tag):
            if x is None: return x
            if x < 0: 
                x += n
            if x < 0 or x >= n:
                print(f'warning: out of bounds {tag} {x}')
                x %= n
            return x
        
        if utterance is not None:
            if start is not None:
                start = self.utterance_to_global_step(utterance, start)
            if end is not None:
                end = self.utterance_to_global_step(utterance, end)
        
        if start is None: start = self.sampler_loop_start        
        if end is None: end = self.sampler_loop_end

        # utterance = wrap(utterance, len(self.states), 'utterance')
        start = wrap(start, len(self.states[utterance]), 'loop start')
        end = wrap(end, len(self.states[utterance]), 'loop end')

        # changed = utterance != self.sampler_utterance
        # self.sampler_utterance = utterance

        self.sampler_loop_start = start
        self.sampler_loop_end = end

        # must reset if changing utterance
        if reset:
            self.reset_sampler()

    def set_sampler_loop_text(self, 
            text:str, n:int=-1, 
            start:bool=True, end:bool=True, 
            reset:bool=True):
        """
        Args:
            text: a regular expression string
            n: index of occurrence in history
            start: if True, set the loop start
            end: if True, set the loop end
            reset: if True, immediately go to loop start
        """
        if not (start or end): return
        r = text

        # get all occurences of all matching strings,
        #   and index by order in history

        # first find matches for the regex in utterance texts
        matches = []
        utts = self.states #if n>=0 else reversed(self.states)
        index = 0
        for utt_index, utt in enumerate(utts):
            if len(utt)==0: continue
            # text = utt[0]['text']
            text = utt.text
            for match in re.findall(r, text):
                # error if multiple capture groups in regex
                if not isinstance(match, str):
                    raise ValueError("`set_sampler_loop`: multiple capture groups not supported")
                # add utterance, text index of match
                index = index + text[index:].find(match)
                matches.append((utt_index, index, index+len(match), match))
                # print(f'{utt_index=}, {index=}, {match=}')

        if len(matches)==0:
            print(f'warning: no text matching "{r}" in `set_sampler_loop`')
            return
        
        # for matches, get occurences and flatten into list
        #   simple: occurrence is from entering start to exiting end
        #   TODO better: occurence is when there is a run which
        #       enters the first token / leaves the last token
        #       without any major jumps or reversals
        #   TODO?: allow skips of first / last token?
        occurrences = []
        for u, i, j, m in matches:
            occ_start = None
            utt = self.states[u]
            for k, state in enumerate(utt):
                align = state['align_hard']
                # print(align)

                if align['enter_index']==i:
                    # open/replace occurrence
                    if occ_start is not None:
                        print(f'warning: double open {m=} {occ_start=} {u=} {k=}')
                    occ_start = k
                if align['leave_index']==j:
                    # close occurrence (or ignore)
                    if occ_start is None: 
                        print(f'warning: close without open {m=}')
                        continue
                    occurrences.append((u, occ_start, k))
                    # print(f'{(u, occ_start, k)=}')
                    occ_start = None

        if len(occurrences)==0:
            print(f'warning: no occurrences of {matches} in `set_sampler_loop`')
            return
        
        # index into list using n
        u, loop_start, loop_end = occurrences[n]

        loop_start = self.utterance_to_global_step(u, loop_start)
        loop_end = self.utterance_to_global_step(u, loop_end)
        # TODO: efficient search for index of occurence
                
        # set loop points
        # if utterance changes, unset points which aren't set
        # changed = u != self.sampler_utterance
        # self.sampler_utterance = u

        if start:
            self.sampler_loop_start = loop_start
        # elif changed:
            # self.sampler_loop_start = 0

        if end:
            self.sampler_loop_end = loop_end
        # elif changed:
            # self.sampler_loop_end = None

        # must reset if changing utterance
        # if changed or reset:
        if reset:
            self.reset_sampler()

        # print(f'{self.sampler_utterance=}, {self.sampler_step=}, {self.sampler_loop_start=}, {self.sampler_loop_end=}')
    
            
    def reset_sampler(self):
        self.step_mode = StepMode.SAMPLER
        self.sampler_step = self.sampler_loop_start
        

    def process_frame(self, 
            latent_t:Optional[Tensor]=None, 
            align_t:Optional[Tensor]=None
        ) -> tuple[Tensor, Tensor]:
        """
        Generate an audio(rave latents) and alignment frame.

        Args:
            latent_t: [batch, RAVE latent size]
                last frame of audio feature if using `latent feedback` mode
            align_t: [batch, text length in tokens]
                explicit alignments if using `paint` mode
        Returns:
            latent_t: [batch, RAVE latent size]
                next frame of audio feature
            align_t: [batch, text length in tokens]
                alignments to text
        """
        with torch.inference_mode():
            with self.profile('tts', detail=False):
                r = self.model.step(
                    alignment=align_t, audio_frame=latent_t, 
                    temperature=self.temperature)
                latent_t, align_t, 
        # use low precision for storage
        return r['output'].half(), r['alignment'].half()
    
    def set_temperature(self, t):
        self.temperature = t
    
    def do_audio_block(self, outdata, sdtime):
        """loop over samples of output, requesting frames from the audio thread
        and pulling them from the queue as needed
        """
        outdata[:,:] = 0
        c = self.synth_channels+self.latent_channels
        for i in range(outdata.shape[0]):
            # if the current frame is exhausted, delete it
            if self.active_frame is not None:
                if self.frame_counter >= self.active_frame.shape[-1]:
                    self.active_frame = None
                    self.frame_counter = 0

            # if no active frame, try to get one from step thread
            if self.active_frame is None:
                if not self.audio_q.empty(): 
                    self.active_frame = self.audio_q.get()

            if not self.trigger_q.full():
                # use ADC input time as timestamp
                timestamp = sdtime.inputBufferAdcTime
                self.trigger_q.put(timestamp)

            if self.active_frame is None:
                if self.playing:
                    print(f'audio: dropped frame')
                return
            else:
                # read next audio sample out of active model frame 
                # TODO: batch/multichannel handling
                outdata[i,c] = self.active_frame[:, self.frame_counter]
                self.frame_counter += 1

    def audio_callback(self,*a):
        """sounddevice callback main loop"""
        if len(a)==4: # output device only case
            (
                outdata,#: np.ndarray, #[frames x channels]
                frames, sdtime, status
            ) = a
        elif len(a)==5: # input and output device
            (
                indata, outdata, #np.ndarray, #[frames x channels]
                frames, sdtime, status
            ) = a

        self.do_audio_block(outdata, sdtime)
         
    def utterance_empty(self, utterance=-1):
        try:
            return len(self.states[utterance])==0
        except IndexError:
            return True
        
    def utterance_len(self, utterance=-1):
        return len(self.states[utterance])
    
    def global_step_to_utterance(self, global_step):
        utterance = 0
        step = global_step
        while True:
            if utterance >= len(self.states):
                raise ValueError('warning: global_step out of bounds')
            ul = len(self.states[utterance])
            if step < ul:
                break
            step -= ul
            utterance += 1

        return utterance, step
    
    def utterance_to_global_step(self, utterance, step):
        if utterance >= len(self.states):
            raise ValueError('warning: utterance out of bounds')
        return step + sum(len(self.states[i]) for i in range(utterance-1))

    def prev_state(self, *a, step=-1, utterance=-1, global_step=None):
        """convenient access to previous states"""
        if global_step is not None:
            try:
                utterance, step = self.global_step_to_utterance(global_step)
            except ValueError:
                return None

        s = self.states[utterance][step]
        for k in a:
            s = s[k]
        return s
    
    def total_steps(self):
        return sum(len(s) for s in self.states)

    def hard_alignment(self, align_t):
        """return index and character of hard alignment"""
        i = align_t.argmax().item()
        # print(i)
        c = self.text[i] if i < len(self.text) else None

        i_enter = None
        i_leave = None
        c_enter = None
        c_leave = None
        if not self.utterance_empty():
            # print(self.states)
            prev_align = self.prev_state('align_hard')

            if i>prev_align['index']:
                i_enter = i
                i_leave = prev_align['index']
                c_enter = c
                c_leave = prev_align['char']

        return {
            'index':i, 
            'char':c, 
            'enter_index':i_enter,
            'leave_index':i_leave,
            'enter_char':c_enter,
            'leave_char':c_leave
        }
    
    def do_reset(self):
        """perform reset of model states/text (called from `step`)"""
        print('RESET')
        for k,v in self.reset_values.items():
            setattr(self, k, v)
        
        if self.text_rep is not None:
            self.model.reset(self.text_rep)
            if self.align_params is None:
                # go to loop start after reset
                self.momentary_align_params = (self.gen_loop_start, 1)
            elif not self.utterance_empty():
                # unless painting alignments -- then try to stay
                # in approximately the same spot
                # TODO this doesn't work because the frontend just sets it again based on the slider -- need to add bidirectional control
                self.align_params = (
                    self.text_index_map(round(self.align_params[0])), 1)
                
            # remove old RNN states
            if len(self.states):
                for state in self.states[-1]:
                    self.strip_states(state)

            # start a new utterance
            self.states.append(Utterance(self.text))
            ## for now, only sampler of current utterance is supported
            self.sampler_step = 0
            # in the future, it might be useful to encode texts without yet starting a new utterance. for now, it makes more sense if hitting encode 
            # always starts generation
            self.generate()

            self.needs_reset = False


        return {
            'reset': True, 
            'text': self.text, 
            'num_latents': self.num_latents, 
            'use_pitch': self.use_pitch
            }

    def paint_alignment(self):
        """helper for `step_gen`"""
        align_params = self.momentary_align_params or self.align_params
        self.momentary_align_params = None
        if align_params is None:
            return None
        loc, scale = align_params
        n_tokens = self.text_rep.shape[1]
        loc = max(0, min(n_tokens, loc))
        deltas = torch.arange(n_tokens) - loc
        deltas = deltas / (0.5 + scale) # sharpness modulated by speed 
        # discrete gaussian, sums exactly to 1 over text positions
        logits = -deltas**2
        res = logits.exp() 
        res = res / res.sum()
        return res[None]
    

    def convert_numpy(self, item):
        if hasattr(item, 'numpy'):
            return item.numpy()
        # elif isinstance(item, dict):
            # return {k:self.convert_numpy(v) for k,v in item.items()}
        # elif hasattr(item, '__len__'):
            # return [self.convert_numpy(i) for i in item]
        else:
            return item
    
    def send_state(self, state):   
        state = {
            k:self.convert_numpy(v)
            for k,v in state.items()
            if k!='model_state'
        }
        self.frontend_conn.send(state)

    def step(self, timestamp):
        """compute one vocoder frame of generation or sampler"""
        # if self.frontend_q.qsize() > 100:
            # self.frontend_q.get()
            # print('frontend queue full, dropping')

        state = {'text':self.text}

        # reset text and model states
        if self.needs_reset:
            state |= self.do_reset()

        # print(f'{self.frontend_conn=} {threading.get_native_id()=} {os.getpid()=}')

        if self.text_rep is None or self.step_mode==StepMode.PAUSE:
            if len(state):
                # self.frontend_q.put(state)
                if self.frontend_conn is not None:
                    self.send_state(state)
                else:
                    print('warning: frontend_conn does not exist')
            return None
        if self.step_mode==StepMode.GENERATION:
            # with self.profile('step_gen'):
            state |= self.step_gen(timestamp)
        elif self.step_mode==StepMode.SAMPLER:
            with self.profile('step_sampler'):
                state |= self.step_sampler(timestamp)
        # else:
            # print(f'WARNING: {self.step_mode=} in step')

        if self.osc_sender is not None:
            self.osc_sender(state)

        if len(self.latent_biases):
            with torch.inference_mode():
                bias = torch.tensor(self.latent_biases)[None]
                state['latent_t'] += bias

        # send to frontend
        # self.frontend_q.put(state)
        if self.frontend_conn is not None:
            self.send_state(state)
        else:
            # print(state)
            print('warning: frontend_conn does not exist')

        latent = state['latent_t'] # batch, channel
        batch, n_latent = latent.shape
        assert batch==1

        c_synth, c_latent = len(self.synth_channels), len(self.latent_channels)

        with torch.inference_mode():
            audio_frame = torch.zeros(
                c_synth+c_latent, 
                self.model.block_size) # channel, time
            
            if OutMode.SYNTH_AUDIO in self.out_mode:
                # allow creative use of vocoders with mismatched sizes
                rave_dim = self.rave.decode_params[0]
                common_dim = min(rave_dim, latent.shape[-1])
                latent_to_rave = torch.zeros(latent.shape[0], rave_dim)
                latent_to_rave[:,:common_dim] = latent[:,:common_dim]
                ### run the vocoder
                with self.profile('vocoder'):
                    audio = self.rave.decode(
                        latent_to_rave[...,None])[0] # channel, time
                ###
                audio_frame[:c_synth, :] = audio
                
            if OutMode.LATENT_AUDIO in self.out_mode:
                latent = torch.cat((
                    torch.zeros(batch, 1), # zero to ensure trigger
                    torch.full((batch, 1), n_latent), # trigger, number of latents
                    latent
                    ), dim=1) / 1024 # scale down in case audio gets sent to speakers
                # latent x batch

                audio_frame[c_synth:, :2+n_latent] = latent

        return audio_frame
        # send to audio thread
        # print(f'DEBUG: queuing audio {id(audio)}')
        # self.audio_q.put(audio_frame)

    def step_sampler(self, timestamp):
        """sampler branch of main `step` method"""
        # print(f'{self.sampler_step=}')

        # if self.utterance_empty(utterance=self.sampler_utterance):
        if self.total_steps()==0:
            print('nothing to play back')
            self.step_mode = StepMode.PAUSE
            return None
        
        state = self.prev_state(global_step=self.sampler_step) or {}

        self.sampler_step += 1

        if (
            self.sampler_step == self.sampler_loop_end 
            or self.sampler_step >= self.total_steps()
            ):
            self.sampler_step = self.sampler_loop_start
            if self.sampler_stop_at_end:
                self.step_mode = StepMode.PAUSE


        state = {**state}
        state['latent_t'] = state['latent_t'].clone()
        state['timestamp'] = timestamp
        state['sampler'] = True

        return state
    

    def step_gen(self, timestamp):
        """generation branch of main `step` method"""
        # loop end
        if (
            self.align_params is None 
            and self.gen_loop_end is not None 
            and not self.utterance_empty()
            and self.prev_state('align_hard', 'index') in (
                self.gen_loop_end, self.gen_loop_end+1)
        ):
            self.momentary_align_params = (self.gen_loop_start, 1)

        # set just model states
        # TODO: possibility to set to previous utterance?
        if self.step_to_set is not None:
            step = self.step_to_set
            with torch.inference_mode():
                try:
                    self.model.set_state(
                        self.prev_state('model_state', step=step))
                except Exception:
                    print(f'WARNING: failed to set {step=} ')
                    raise
            self.step_to_set = None

        if self.model.memory is None:
            print('skipping: model not initialized')
            if self.rave is not None:
                self.audio_q.put(None)
            return None

        if not self.utterance_empty() and self.latent_feedback:
            # pass the last vocoder frame back in
            latent_t = self.prev_state('latent_t')
        else:
            latent_t = None

        align_t = self.paint_alignment()

        ##### run the TTS model
        latent_t, align_t = self.process_frame(
            # self.mode, 
            latent_t=latent_t, align_t=align_t)
        #####

        # NOTE
        ### EXPERIMENTAL juice knob
        # latent_t = latent_t * 1.1
        ###

        state = {
            # 'text':self.text,
            'latent_t':latent_t, 
            'align_t':align_t,
            'timestamp':timestamp,
            'model_state':self.model.get_state(),
            'align_hard':self.hard_alignment(align_t)
            }
        
        if (
            self.generate_stop_at_end 
            and state['align_hard']['index']==(len(self.text)-1)
        ):
            self.step_mode = StepMode.PAUSE
            # self.reset()

        utt = self.states[-1]
        utt.append(state)
        # remove RNN states after certain number of steps
        if len(utt) >= self.max_model_state_storage:
            self.strip_states(utt[-self.max_model_state_storage])
        return state
    
    def strip_states(self, state):
        """remove RNN states"""
        state.pop('model_state', None)

    @property
    def playing(self):
        return self.step_mode!=StepMode.PAUSE


def main(checkpoint='../rtalign_004_0100.ckpt'):
    from tungnaa.gui.senders import DirectOSCSender
    b = Backend(checkpoint, osc_sender=DirectOSCSender())
    b.start()
    print('setting text')
    n = b.set_text('hey, hi, hello, test. text. this is a text.') # returns the token length of the text
    print(f'text length is {n} tokens')
    print('setting alignment')
    i = 0
    while True:
        b.set_alignment(torch.randn(1,n).softmax(-1)) # ignored if mode == paint
        if i==300:
            print('setting text')
            n = b.set_text('this is a new text')
        if i==500:
            print('setting mode')
            b.set_mode('paint') # the other available mode is 'infer'

        while not b.q.empty():
            r = b.q.get()
            a = r['align_t'].argmax().item()
            print('█'*(a+1) + ' '*(n-a-1) + '▏')

            print(' '.join(f'{x.item():+0.1f}' for x in r['latent_t'][0]))
            
        time.sleep(1e-2)
        i += 1

if __name__=='__main__':
    fire.Fire(main)



import os
# fix torch device order to be same as nvidia-smi order
os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"

from pathlib import Path
import random
from collections import defaultdict
import itertools as it
from typing import Tuple

from tqdm import tqdm
import fire

import numpy as np
import matplotlib.pyplot as plt

import torch
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

torch.set_float32_matmul_precision('high')
torch.backends.cudnn.benchmark = True

from tungnaa.model import TacotronDecoder
from tungnaa.util import deep_update, get_class_defaults, ConcatSpeakers
from tungnaa.split import get_datasets

class Trainer:
    """Instantiate a Trainer object

    follow with a `train` subcommand to start training

    Args:
        experiment: experiment name
        model_dir: where to store checkpoints
        log_dir: where to store tensorboard logs
        manifest: path to HiFiTTS-style json manifest
        rave_model: path to vocoder .ts file, for logs and determining latent size
        csv: path to additional annotations in a CSV file
            currently based on the format of `jvs_labels_encoder_k7.csv`
            the first column should be the name of the audio file without extension
            the second column will be added to the text, separated by a colon:
            "val:original text"
        concat_speakers: number of utterances to concatenate.
            for each training example, this will load n utterances from the dataset,
            apply any annotations from `csv` or `speaker_annotate`,
            then concatenate the texts and audio.
        strip_quotes: remove all double quotes from text
        speaker_annotate: prepend speaker id to text, as "[speaker]"
        speaker_dataset: when speaker_annotate is True,
            also prepend speaker dataset id to text, as "[dataset:speaker]"
        results_dir: where to store results
        model: dict of model constructor overrides
            e.g. '{text_encoder:{end_tokens:True}, likelihood_type:nsf, flow_blocks:2, nsf_bins:16, prenet_dropout:0.2, rnn_size:1200, decoder_layers:1, frame_channels:11}'
        freeze_text: freeze text encoder
        freeze_embeddings: freeze text encoder embeddings only
        init_text: reinitialize pretrained text encoder
        batch_size: training batch dimensions
        batch_max_tokens: max text length during training, in tokens.
            should be set to usually result in longer text than audio.
            (this depends on the vododer rate and tokens/second of the speech)
        batch_max_frames: max audio length during training, in vocoder frames.
            should be set to usually result in longer text than audio.
            (this depends on the vododer rate and tokens/second of the speech)
        lr: learning rate
        lr_text: override learning rate for text encoder
        adam_betas: AdamW optimizer beta parameters
        adam_eps: AdamW optimizer epsilon parameter
        weight_decay: AdamW optimizer weight decay parameter
        grad_clip: gradient clipping
        seed: random seed
        n_jobs: dataloader parallelism
        device: training device, e.g. 'cpu', 'cuda:0'
        epoch_size: in iterations, None for whole dataset
        valid_size: set size of validation set, in batches
            may be useful if validation set is very large or small
        save_epochs: save checkpoint every so many epochs
        nll_scale: scale NLL loss
        data_portion: fraction of dataset to train on
            (for restricted data experiments)
        debug_loading: debug dataloading
        jit: attempt to compile model with torch.jit.script
        checkpoint: file for resuming training or transfer learning
        resume: if True, resume training, otherwise transfer weights
        rand_text_subs: dict of text substitutions
        style_annotate: probability of prepending style annotations
    """
    def __init__(self, 
        experiment:str, 
        model_dir:Path,
        log_dir:Path, 
        manifest:Path,
        rave_model:Path = None,
        csv:Path=None,
        concat_speakers:int = 0,
        strip_quotes:bool = True,
        speaker_annotate:bool = False,
        speaker_dataset:bool = False,
        results_dir:Path = None,
        model = None, 
        freeze_text:bool = False,
        freeze_embeddings:bool = True,
        init_text:bool = False, 
        batch_size:int = 32,
        # TODO: specify estimated tokens / frame ? 
        batch_max_tokens:int = 256,
        batch_max_frames:int = 512, 
        lr:float = 3e-4,
        lr_text:float = None,
        adam_betas:Tuple[float,float] = (0.9, 0.998),
        adam_eps:float = 1e-08, 
        weight_decay:float = 1e-6,
        grad_clip:float = 5.0,
        seed:int = 0, 
        n_jobs:int = 4,
        device:str = 'cuda:0',
        epoch_size:int = None,
        valid_size:int = None,
        save_epochs:int = 1,
        max_epochs:int = None,
        nll_scale:float = 1, 
        dispersion_scale:float = 0, 
        dispersion_cutoff:float = 0,
        concentration_scale:float = 0, 
        concentration_norm_scale:float = 0, 
        concentration_cutoff:float = 0, 
        # dispersion_l2_scale:float = 0, 
        # concentration_l2_scale:float = 0, 
        # TODO: anneal_prenet = None, # number of epochs to anneal prenet dropout to zero
        data_portion:float = 1,
        debug_loading:bool = False,
        jit:bool = False,
        compile:bool = False,
        checkpoint:Path = None,
        resume:bool = True,
        rand_text_subs = None,
        replace_runs = False,
        style_annotate = 0,
        drop_prefix = None
        ):

        kw = dict(locals()); kw.pop('self')

        kw.pop('checkpoint')
        if checkpoint is not None:
            print(f'loading checkpoint {checkpoint}')
            checkpoint = torch.load(checkpoint, map_location=torch.device('cpu'))
            # merges sub dicts, e.g. model hyperparameters
            deep_update(checkpoint['kw'], kw) # provided arguments override stored
            kw = checkpoint['kw']
            load = lambda: self.load_state(checkpoint, resume=resume)
        else:
            load = lambda: None

        # store all hyperparams for checkpointing
        self.kw = kw

        self.best_step = None
        self.best_loss = np.inf

        # get model defaults from model class
        model_cls = TacotronDecoder
        if model is None: model = {}
        assert isinstance(model, dict), """
            model keywords are not a dict. check shell/fire syntax
            """
        kw['model'] = model = get_class_defaults(model_cls) | model

        # assign all arguments to self by default
        self.__dict__.update(kw)
        # mutate some arguments:
        # self.lr_text = self.lr_text or self.lr
        self.model_dir = Path(model_dir) / self.experiment
        self.log_dir = Path(log_dir) / self.experiment
        if results_dir is None:
            self.results_dir = None
        else:
            self.results_dir = Path(results_dir) / self.experiment
        self.manifest = Path(manifest)
        self.device = torch.device(device)

        # filesystem
        for di in (self.model_dir, self.log_dir, self.results_dir):
            if di is not None:
                di.mkdir(parents=True, exist_ok=True)

        if rave_model is None:
            self.rave_model = None
        else:
            self.rave_model = torch.jit.load(rave_model)
            model['block_size'] = self.rave_model.encode_params[3]
            model['frame_channels'] = self.rave_model.latent_size
            
        # random states
        self.seed_random()

        # logging
        self.writer = SummaryWriter(self.log_dir)

        # Trainer state
        self.iteration = 0
        self.exposure = 0
        self.epoch = 0

        if jit:
            kw['model']['script'] = True

        # construct model from arguments 
        self.model = model_cls(**model)
        tqdm.write(repr(self.model))

        tqdm.write(f'{sum(p.numel() for p in self.model.parameters())} parameters')

        if init_text:
            self.model.text_encoder.init()
        if freeze_text:
            self.model.text_encoder.requires_grad_(False)
        if freeze_embeddings:
            for n,m in self.model.text_encoder.named_modules():
                if isinstance(m, torch.nn.Embedding):
                    tqdm.write(f'freezing {n}')
                    m.requires_grad_(False)

        tqdm.write(f'{sum(p.numel() for p in self.model.parameters() if p.requires_grad)} trainable parameters')

        tqdm.write(f'moving model to {self.device}')
        self.model = self.model.to(self.device)

        try:
            self.model.core.attention_rnn.flatten_parameters()
        except Exception:
            tqdm.write('did not flatten attention_rnn parameters')
        try:
            self.model.core.decoder_rnn.flatten_parameters()
            tqdm.write('did not flatten decoder_rnn parameters')
        except Exception:
            pass
      
        # if jit:
            # print(f'scripting model')
            # self.model = torch.jit.script(self.model)
            # self.model.core_loop = torch.jit.script(self.model.core_loop)

        if compile:
            tqdm.write(f'compiling model')
            self.model.core.compile(
                dynamic=True, 
                mode='reduce-overhead'
                )
            # self.model.compile(
            #     dynamic=True, 
            #     mode='reduce-overhead'
            #     )

        tqdm.write(f'constructing datasets')
        self.train_dataset, self.valid_dataset, self.test_dataset = get_datasets(
            manifest, 
            data_portion=data_portion, 
            # concat_speakers=concat_speakers,
            csv_file=csv, 
            max_tokens=batch_max_tokens, 
            max_frames=batch_max_frames,
            speaker_annotate=speaker_annotate,
            speaker_dataset=speaker_dataset,
            strip_quotes=strip_quotes, 
            rave=self.rave_model,
            text_encoder=self.model.text_encoder,
            rand_text_subs=rand_text_subs,
            style_annotate=style_annotate,
            replace_runs=replace_runs
        )
        for tag, ds in (('train', self.train_dataset), ('valid', self.valid_dataset)):
            time_s = 0
            text_c = 0
            for item in ds:
                time_s += item['audio'].shape[1] * self.rave_model.encode_params[3] / self.rave_model.sr
                text_c += len(item['plain_text'])
            tqdm.write(f'{tag} dataset length: {len(ds)} utterances, {time_s} seconds, {text_c} characters')

        self.valid_size = valid_size or len(self.valid_dataset)//batch_size

        if concat_speakers > 1:
            self.train_dataset = ConcatSpeakers(self.train_dataset, concat_speakers, drop_prefix=drop_prefix)
            self.valid_dataset = ConcatSpeakers(self.valid_dataset, concat_speakers, drop_prefix=drop_prefix)
            self.test_dataset = ConcatSpeakers(self.test_dataset, concat_speakers, drop_prefix=drop_prefix)

            
        tqdm.write(f'creating optimizer')
        if self.lr_text is None:
            self.lr_text = self.lr
        params = [{
            'params':p, 'lr':(self.lr_text if 'text_encoder' in n else self.lr)
        } for n, p in self.model.named_parameters()]

        self.opt = torch.optim.AdamW(params,
            self.lr, self.adam_betas, self.adam_eps, self.weight_decay)
        
        load()
        

    @property
    def gpu(self):
        return self.device.type!='cpu'

    def seed_random(self):
        random.seed(self.seed)
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)

    def set_random_state(self, states):
        # note: GPU rng state not handled
        std_state, np_state, torch_state = states
        random.setstate(std_state)
        np.random.set_state(np_state)
        torch.set_rng_state(torch_state)

    @property
    def step(self):
        return self.exposure, self.iteration, self.epoch

    def save(self, fname):
        torch.save(dict(
            kw=self.kw,
            model_state=self.model.state_dict(),
            optimizer_state=self.opt.state_dict(),
            step=self.step,
            best_loss=self.best_loss,
            best_step=self.best_step,
            random_state=(random.getstate(), np.random.get_state(), torch.get_rng_state())
        ), fname)

    def load_state(self, d, resume):
        print(f'{resume=}')
        d = d if hasattr(d, '__getitem__') else torch.load(d)
        sd = d['model_state']
        sd = self.model.update_state_dict(sd)
        if not resume:
            msd = self.model.state_dict()
            for k in list(sd):
                print(k, sd[k].shape, msd[k].shape)
                if k not in msd or sd[k].shape != msd[k].shape:
                    print(f'skipping {k}')
                    sd.pop(k)
        self.model.load_state_dict(sd, strict=resume)
        # self.model.load_state_dict(d['model_state'], strict=resume)
        if resume:
            print('loading optimizer state, RNG state, step counts')
            print("""
            warning: optimizer lr, beta etc are restored with optimizer state,
            even if different values given on the command line, when resume=True
            """)
            self.opt.load_state_dict(d['optimizer_state'])
            self.exposure, self.iteration, self.epoch = d['step']
            self.set_random_state(d['random_state'])
            try:
                self.best_loss = d['best_loss']
                self.best_step = d['best_step']
            except KeyError:
                print('old checkpoint: no best_loss')
        else:
            print('fresh run transferring only model weights')

    def log(self, tag, d):
        # self.writer.add_scalars(tag, d, self.exposure)
        for k,v in d.items():
            self.writer.add_scalar(f'{tag}/{k}', v, self.exposure)
    
    def process_grad(self):
        r = {}
        if self.grad_clip is not None:
            r['grad_l2'] = torch.nn.utils.clip_grad_norm_(
                self.model.parameters(), self.grad_clip, error_if_nonfinite=True)
        return r

    def get_loss_components(self, result):
        # return {'error': result['mse']}
        return {
            'nll': result['nll'].mean()*self.nll_scale,
            'dispersion': (
                result['dispersion']
                    .clip(self.dispersion_cutoff, 1e9)
                    .mean()
                    *self.dispersion_scale),
            'concentration': (
                result['concentration']
                    .mean()
                    *self.concentration_scale),
            'concentration_norm': (
                result['concentration_norm']
                    .clip(self.concentration_cutoff, 1e9)
                    .mean()
                    *self.concentration_norm_scale),
            # 'dispersion_l2': result['dispersion_l2'].mean()*self.dispersion_l2_scale,
            # 'concentration_l2': result['concentration_l2'].mean()*self.concentration_l2_scale,
            }

    def forward(self, batch):
        audio_lengths = batch['audio_mask'].sum(-1).cpu()
        text_mask = batch['text_mask'].to(self.device, non_blocking=True)
        audio_mask = batch['audio_mask'].to(self.device, non_blocking=True)
        if self.model.text_encoder is not None:
            text = batch['text'].to(self.device, non_blocking=True)
        else:
            text = batch['text_emb']
            if text is None:
                raise ValueError("""
                    no text embeddings in dataset but no text encoder in model
                """)
            text = text.to(self.device, non_blocking=True)
        audio = batch['audio'].to(self.device, non_blocking=True)
        # torch.compiler.cudagraph_mark_step_begin()
        return self.model(
            text, audio, text_mask, audio_mask, audio_lengths, 
            chunk_pad_text=128 if self.compile else None)

    def get_scalars(self, d):
        r = {}
        for k,v in d.items():
            if v.numel()==1:
                r[k] = v.item()
            elif v.shape==(self.batch_size,):
                r[k] = v.mean().item()
        return r

    def _validate(self, valid_loader, ar_mask=None):
        """"""
        pops = defaultdict(list)
        self.model.eval()
        i = 0
        # for batch in tqdm(valid_loader, desc=f'validating epoch {self.epoch}'):
        vs = self.valid_size
        for batch in tqdm(
                it.islice(it.chain.from_iterable(it.repeat(valid_loader)), vs), 
                desc=f'validating epoch {self.epoch}', total=vs):
            with torch.no_grad():
                result = self.forward(batch)
                losses = {
                    k:v.item() 
                    for k,v in self.get_loss_components(result).items()}
                for k,v in losses.items():
                    pops[k].append(v)
                pops['loss'].append(sum(losses.values()))
                # pops['mse'].append(result['mse'].item())
                # pops['nll_fixed'].append(result['nll_fixed'].item())
                # pops['nll_posthoc'].append(result['nll_posthoc'].item())
                for k,v in self.get_scalars(result).items():
                    pops[k].append(v)
                # for k,v in result.items():
                    # NOTE: here (and in train loop), could check for batch size vector and take mean
                    # that would allow processing losses per batch item in training script
                    # if v.numel()==1:
                        # pops[k].append(v.item())
            if i==0:
                self.rich_logs('valid', batch, result)
                i+=1
        return {
            'logs':{k:np.mean(v) for k,v in pops.items()},
            'pops':pops
        }

    def train(self):
        """Entry point to model training"""
        try:
            self._train()
        except Exception as e:
            import traceback; traceback.print_exc()
            import pdb; pdb.post_mortem()

    def _train(self):
        # self.save(self.model_dir / f'{self.epoch:04d}.ckpt')

        # TODO: can remove this?
        def get_collate(ds):
            while True:
                try:
                    return ds.collate_fn()
                except AttributeError:
                    ds = ds.dataset

        train_loader = DataLoader(
            self.train_dataset, self.batch_size,
            shuffle=not isinstance(
                self.train_dataset, torch.utils.data.IterableDataset),
            num_workers=self.n_jobs, pin_memory=self.gpu,
            worker_init_fn=getattr(self.train_dataset, 'worker_init', None),
            persistent_workers=True,
            collate_fn=get_collate(self.train_dataset),
            drop_last=True)

        valid_loader = DataLoader(
            self.valid_dataset, self.batch_size,
            shuffle=False, num_workers=self.n_jobs, pin_memory=self.gpu,
            worker_init_fn=getattr(self.valid_dataset, 'worker_init', None),
            persistent_workers=False,
            collate_fn=get_collate(self.valid_dataset))

        ##### validation loop
        def run_validation():
            if self.debug_loading: return
            logs = self._validate(valid_loader)['logs']
            self.log('valid', logs)
            if logs['loss'] < self.best_loss:
                self.best_loss = logs['loss']
                self.best_step = self.step
                self.save(self.model_dir / f'best.ckpt')

        try:
            epoch_size = self.epoch_size or len(train_loader)
        except TypeError:
            raise ValueError("specify epoch_size when using IterableDataset")

        # validate at initialization
        run_validation()
        
        if self.debug_loading:
            f = open('debug.txt', 'w')

        while self.max_epochs is None or self.epoch < self.max_epochs:
            self.epoch += 1

            ##### training loop
            self.model.train()
            for batch in tqdm(
                # itertools incantation to support epoch_size larger than train set
                # NOTE important to use persistent_workers in DataLoader!
                # otherwise RNG gets repeated
                it.islice(
                    it.chain.from_iterable(it.repeat(train_loader)), epoch_size), 
                desc=f'training epoch {self.epoch}', total=epoch_size
                ):

                self.iteration += 1
                self.exposure += self.batch_size
                logs = {}

                if self.debug_loading:
                    f.write(f'{self.epoch}, {self.iteration}, {list(batch["index"].numpy())}\n')
                    continue

                ### forward+backward+optimizer step ###
                self.opt.zero_grad(set_to_none=True)
                result = self.forward(batch)
                losses = self.get_loss_components(result)
                loss = sum(losses.values())
                loss.backward()
                logs |= self.process_grad()
                self.opt.step()
                ########

                # log loss components
                logs |= {f'loss/{k}':v.item() for k,v in losses.items()}
                # log total loss
                logs |= {'loss':loss.item()}
                # log any other returned scalars
                # logs |= {k:v.item() for k,v in result.items() if v.numel()==1}
                logs |= self.get_scalars(result)
                # other logs
                self.log('train', logs)

            if self.epoch%self.save_epochs == 0: 
                self.save(self.model_dir / f'{self.epoch:04d}.ckpt')
            self.save(self.model_dir / f'last.ckpt')
            run_validation()


    def add_audio(self, tag, audio):
        try:
            sr = self.rave_model.sampling_rate
        except Exception:
            sr = self.rave_model.sr
        self.writer.add_audio(
            tag, audio, 
            global_step=self.epoch, sample_rate=int(sr))

    def wrap_text(self, s, n):
        w = ''
        while s!='':
            w = w+'\n'+s[:n]
            s = s[n:]
        return w

    def rich_logs(self, tag, batch, result):
        """runs RAVE inference and logs audio"""
        z = result['predicted'].detach().cpu()
        gt = result['ground_truth'].detach().cpu()
        a = result['alignment'].detach().cpu()
        am = result['audio_mask'].detach().cpu()
        tm = result['text_mask'].detach().cpu()
        for i in tqdm(range(3), desc='rich logs', leave=False):
            nt = tm[i].sum()
            na = am[i].sum()
            z_i = z[i, :na].T[None]
            gt_i = gt[i, :na].T[None]
            with torch.inference_mode():
                audio_tf = self.rave_model.decode(z_i)[0]
                audio_gt = self.rave_model.decode(gt_i)[0]
            self.add_audio(f'{tag}/audio/tf/{i}', audio_tf)
            self.add_audio(f'{tag}/audio/gt/{i}', audio_gt)
            
            with torch.inference_mode():
                t = result['text'].detach()[i:i+1, :nt]
                z_ar,_ = self.model.inference(
                    t, stop=False, max_steps=int(na*1.2))
                z_ar = z_ar.cpu().transpose(1,2)
                z_ar_zero,_ = self.model.inference(
                    t, stop=False, max_steps=int(na*1.2), temperature=0.0)
                z_ar_zero = z_ar_zero.cpu().transpose(1,2)
                z_ar_half,_ = self.model.inference(
                    t, stop=False, max_steps=int(na*1.2), temperature=0.5)
                z_ar_half = z_ar_half.cpu().transpose(1,2)
                audio_ar = self.rave_model.decode(z_ar)[0]
                audio_ar_zero = self.rave_model.decode(z_ar_zero)[0]
                audio_ar_half = self.rave_model.decode(z_ar_half)[0]
            self.add_audio(f'{tag}/audio/ar/{i}', audio_ar)
            self.add_audio(f'{tag}/audio/ar/zerotemp/{i}', audio_ar_zero)
            self.add_audio(f'{tag}/audio/ar/halftemp/{i}', audio_ar_half)

            fig = plt.figure()
            a_i = a[i, :na, :nt].T
            plt.imshow(
                a_i, 
                interpolation='nearest', 
                aspect='auto', 
                origin='lower')
            plt.title(self.wrap_text(batch['plain_text'][i], 40))
            plt.xlabel(self.wrap_text(batch['audio_path'][i], 64))
            plt.tight_layout()
            self.writer.add_figure(f'{tag}/align/{i}', fig, global_step=self.epoch)

    
def resume(checkpoint, resume:bool=True, **kw):
    """
    Args:
        checkpoint: path to training checkpoint file
        resume: if True, restore optimizer states etc
            otherwise, restore only model weights (for transfer learning)
    """
    d = torch.load(checkpoint, map_location=torch.device('cpu'))
    print(f'loaded checkpoint {checkpoint}')
    # merges sub dicts, e.g. model hyperparameters
    deep_update(d['kw'], kw)
    trainer = Trainer(**d['kw'])
    trainer.load_state(d, resume=resume)

    return trainer

# class Resumable:
#     def __init__(self, checkpoint:Path=None, resume:bool=True, **kw):
#         """
#             checkpoint: path to training checkpoint file
#             resume: if True, restore optimizer states etc
#                 otherwise, restore only model weights (for transfer learning)
#         """
#         if checkpoint is not None:
#             d = torch.load(checkpoint, map_location=torch.device('cpu'))
#             print(f'loaded checkpoint {checkpoint}')
#             # merges sub dicts, e.g. model hyperparameters
#             deep_update(d['kw'], kw)
#             self.trainer = Trainer(**d['kw'])
#             self.trainer.load_state(d, resume=resume)
#         else:
#             self.trainer = Trainer(**kw)

#     def train(self):
#     #     import torch._dynamo.config
#     #     torch._dynamo.config.verbose=True
#     #     self._trainer.compile()
#         self.trainer.train()

#     # def test(self):
#     #     self._trainer.test()

# Resumable.__init__.__doc__ = Resumable.__init__.__doc__ + Trainer.__init__.__doc__
# Resumable.train.__doc__ = Trainer.train.__doc__


if __name__=='__main__':
    # TODO: improve fire-generated help message
    try:   
        fire.Fire()
        # fire.Fire(Resumable)
    except Exception as e:
        import traceback; traceback.print_exc()
        import pdb; pdb.post_mortem()


import markdown as md
import huggingface_hub as hf
from pathlib import Path

def get_metadata_files(repo):
    fs = hf.HfFileSystem()
    hf_files = fs.glob(f'{repo}/models/tts/*.md')
    # download and cache each markdown file
    for f in hf_files:
        file = f.split(repo)[-1]
        if file.startswith('/'):
            file = file[1:]
        # print(file)
        yield hf.hf_hub_download(repo_id=repo, filename=file)    

def read_markdown(f):
    with open(f, 'r') as file:
        text = file.read()
    m = md.Markdown(extensions=['full_yaml_metadata'])
    m.convert(text)
    return m

def get_markdown(repo):
    for f in get_metadata_files(repo):
        k = Path(f).stem
        yield (k, read_markdown(f))

def dl_model(repo, tts, vocoder):
    for n,m in get_markdown(repo):
        if n==tts:
            tts_hf = f'models/tts/{tts}.ckpt'
            tts = hf.hf_hub_download(repo_id=repo, filename=tts_hf)
            print(f'{tts=}')
            if vocoder is None:
                voc_hf = f'models/vocoder/{m.Meta["vocoder"]}'
                vocoder = hf.hf_hub_download(repo_id=repo, filename=voc_hf)
                print(f'{vocoder=}')
            return tts, vocoder

def main(repo='Intelligent-Instruments-Lab/tungnaa-models-public'):
    for n,m in get_markdown(repo):
        print(n)
        for k,v in m.Meta.items():
            print(f'\t{k}: {v}')
